// --- Consts
const int C_Invalid = -123;
const uint C_MatchFoundTimeout = 15000;

// === CLASSES ===

// Queue System

class Queue {
	int QueueID;
	string Name;
	QueueStats@ Stats = QueueStats();
	string Status;
	int64 LastPlayerStatusCheck;

	string GetTitle() {
		string Addition;
		if (!(Status == "enabled" || Status == "")) {
			Addition = " \\$999(" + Status + ")";
		}
		return Name + Addition;
	}
}

class QueueTemplate {
	int QueueID;
	string Name;

	QueueTemplate(int id, const string &in name)
	{
		QueueID = id;
		Name = name;
	}
}

class QueueStats {
	QueueRanking@ CurrentRanking = QueueRanking();
	array<StatsMatch@> PlayerMatches;

	array<ProgressionPoint@> RankProgression;
	array<ProgressionPoint@> PointProgression;
	
	array<Division@> Divisions;
	Division@ CurrentDivision;

	int64 LastKnownMatchTime = 0;

	QueueLeaderboard@ Leaderboard = QueueLeaderboard();
	
	Division@ GetDivisionFromRanking(QueueRanking@ Ranking) {
		Division@ div = Divisions[0];

		// Iterate through divisions and check if user is in, always prefer higher division position
		for (uint i = 0; i < Divisions.Length; i++) {
			if (Divisions[i].RuleType == DivisionRuleType::PointsRange) {
				if (Ranking.Points >= Divisions[i].MinPoints && Ranking.Points <= Divisions[i].MaxPoints) {
					if (div is null || Divisions[i].Position > div.Position) @div = Divisions[i];
				}
			} else if (Divisions[i].RuleType == DivisionRuleType::MinimumPoints) {
				if (Ranking.Points >= Divisions[i].MinPoints) {
					if (div is null || Divisions[i].Position > div.Position) @div = Divisions[i];
				}
			} else if (Divisions[i].RuleType == DivisionRuleType::MinimumRankAndPoints) {
				if (Ranking.Points >= Divisions[i].MinPoints && Ranking.Rank >= Divisions[i].MinRank) {
					if (div is null || Divisions[i].Position > div.Position) @div = Divisions[i];
				}
			} else if (Divisions[i].RuleType == DivisionRuleType::VictoryCount) {
				if (Ranking.Wins >= Divisions[i].MinVictories && Ranking.Wins <= Divisions[i].MaxVictories) {
					if (div is null || Divisions[i].Position > div.Position) @div = Divisions[i];
				}
			} else {
				if (Ranking.Rank <= Divisions[i].MinRank) {
					if (div is null || Divisions[i].Position > div.Position) @div = Divisions[i];
				}
			}
		}

		return div;
	}

	void RefreshDivisionRanking() {
		bool CheckRanking = (CurrentDivision !is null);
		int PrevPosition = 0;
		if (CheckRanking) {
			PrevPosition = CurrentDivision.Position;
		}

		@CurrentDivision = GetDivisionFromRanking(CurrentRanking);

		// User is in no division? Create placeholder
		if (CurrentDivision is null) {

			// What went wrong? Let's analyze:
			string LogString = "\nCurrent Points: " + CurrentRanking.Points + "\nCurrent Rank: " + CurrentRanking.Rank + "\nDivisions:\n";
			for (uint i = 0; i < Divisions.Length; i++) {
				LogString += Divisions[i].SerializeString();
			}
			Error(ErrorType::Error, "RefreshDivisionRanking", "Current division Invalid! Creating placeholder.", LogString);
			
			// Placeholder
			Division@ div = Division();
			div.DivisionID = "unknown";
			div.Position = 0;
			div.RuleType = DivisionRuleType::PointsRange;
			div.MinPoints = -100000;
			div.MaxPoints = 100000;
			@div = div;
		}
	}

	void CreateDefaultDivisions(bool Royal = false) {
		Divisions.RemoveRange(0, Divisions.Length);
		if (!Royal) {
			const uint[] BorderPoints = {299, 599, 999, 1299, 1599, 1999, 2299, 2599, 2999, 3299, 3599, 3999, 10000};
			for (uint i = 0; i < BorderPoints.Length; i++) {
				Division@ div = Division();
				div.DivisionID = "DIVISION" + i;
				div.Position = i + 1;
				div.RuleType = DivisionRuleType::PointsRange;
				if (i == 0) {
					div.MinPoints = 0;
				} else {
					div.MinPoints = BorderPoints[i-1] + 1;
				}
				div.MaxPoints = BorderPoints[i];
				Divisions.InsertLast(div);
			}
		} else {
			const uint[] BorderWins = {1, 10, 100, 1000};
			for (uint i = 0; i < BorderWins.Length; i++) {
				Division@ div = Division();
				div.DivisionID = "placeholder";
				div.Position = i + 1;
				div.RuleType = DivisionRuleType::PointsRange;
				if (i == 0) {
					div.MinPoints = 0;
				} else {
					div.MinPoints = BorderWins[i-1] + 1;
				}
				div.MaxPoints = BorderWins[i];
				Divisions.InsertLast(div);
			}
		}
	}

	void SaveMatchForStats(Match@ match) {
		if (match.Status != "COMPLETED") {
			Error(ErrorType::Error, "SaveMatchForStats", "To-be saved match has to be completed!", "");
			return;
		}
		StatsMatch@ MatchSave = StatsMatch();
		int UserTeam = -1;

		// Check match MVP
		string ParticipantsResponse = SendNadeoRequest(Net::HttpMethod::Get, "https://meet.trackmania.nadeo.club/api/matches/" + match.LID + "/participants");
		Json::Value ParticipantsResult = ResponseToJSON(ParticipantsResponse, Json::Type::Array);
		if (ParticipantsResult.GetType() != Json::Type::Null) {
			try {
				for (uint i = 0; i < ParticipantsResult.Length; i++) {
					if (ParticipantsResult[i]["participant"] == GetMainUserInfo().WebServicesUserId) {
						MatchSave.MVP = bool(ParticipantsResult[i]["mvp"]);
						UserTeam = int(ParticipantsResult[i]["teamPosition"]);
						break;
					}
				}
			} catch {
				Error(ErrorType::Error, "SaveMatchForStats", "Something went wrong parsing the match participants result!", getExceptionInfo() + "\n" + ParticipantsResponse);
			}
		} else {
			Error(ErrorType::Error, "SaveMatchForStats", "Couldn't parse match participants result!", ParticipantsResponse);
		}

		// Check Match team result
		string TeamResultResponse = SendNadeoRequest(Net::HttpMethod::Get, "https://meet.trackmania.nadeo.club/api/matches/" + match.LID + "/teams");
		Json::Value TeamResultJSON = ResponseToJSON(TeamResultResponse, Json::Type::Array);
		if (TeamResultJSON.GetType() != Json::Type::Null) {
			try {
				for (uint i = 0; i < TeamResultJSON.Length; i++) {
					if (TeamResultJSON[i]["position"] == UserTeam) {
						MatchSave.TeamResult = TeamResultJSON[i]["rank"];
						break;
					}
				}
			} catch {
				Error(ErrorType::Error, "SaveMatchForStats", "Something went wrong parsing the match teams result!", getExceptionInfo() + "\n" + TeamResultResponse);
			}
		} else {
			Error(ErrorType::Error, "SaveMatchForStats", "Couldn't parse match teams result!", TeamResultResponse);
		}
		
		MatchSave.LID = match.LID;
		MatchSave.Maps = match.Maps;
		MatchSave.TimeStamp = match.StartDate;
		MatchSave.PointsChange = C_Invalid;

		PlayerMatches.InsertAt(0,MatchSave);
	}

	QueueStats() {
		CreateDefaultDivisions();
	}
}

enum DivisionRuleType {
	PointsRange,
	MinimumRank,
	MinimumPoints,
	MinimumRankAndPoints,
	VictoryCount
}

class Division {
	string DivisionID;
	int Position;
	DivisionRuleType RuleType;
	int MinPoints;
	int MaxPoints;
	int MinRank;
	int MinVictories;
	int MaxVictories;

	string SerializeString() {
		string ReturnedValue = "DivisionID: " + DivisionID + "\n";
		ReturnedValue += "Position: " + Position + "\n";
		ReturnedValue += "RuleType: " + RuleType + "\n";
		ReturnedValue += "MinPoints: " + MinPoints + "\n";
		ReturnedValue += "MaxPoints: " + MaxPoints + "\n";
		ReturnedValue += "MinRank: " + MinRank + "\n";
		ReturnedValue += "MinVictories: " + MinVictories + "\n";
		ReturnedValue += "MaxVictories: " + MaxVictories + "\n";
		return ReturnedValue;
	}
}

class ProgressionPoint {
	int64 TimeStamp;
	int DataValue;
}

class QueueRanking {
	int Points = 0;
	int Rank = 0;
	int Wins = 0;
	int Cardinal = 0;
}

class QueueLeaderboard {
	bool NewPageRequested;
	bool LastPageEmpty;
	array<LeaderboardPlayer@> PlayerLeaderboard;
	uint CurrentPage;
	uint64 LastRequestStamp;
	uint FailedRequestsCount = 0;

	void RequestNextPage() {
		LastRequestStamp = Time::Now;

		// Get Leaderboard JSON
		string LeaderboardPageResponse = SendJSONRequest(Net::HttpMethod::Get, "https://trackmania.io/api/top/matchmaking/" + G_QueueState.CurrentQueue.QueueID + "/" + CurrentPage);
		Json::Value LeaderboardPage = ResponseToJSON(LeaderboardPageResponse, Json::Type::Object);
		if (LeaderboardPage.GetType() != Json::Type::Null && LeaderboardPage.HasKey("ranks")) {
			try {
				if (LeaderboardPage["ranks"].Length == 0) LastPageEmpty = true;
				for (uint i = 0; i < LeaderboardPage["ranks"].Length; i++) {
					PlayerLeaderboard.InsertLast(LeaderboardPlayer(LeaderboardPage["ranks"][i]["rank"], LeaderboardPage["ranks"][i]["score"], Player(LeaderboardPage["ranks"][i]["player"]["id"], LeaderboardPage["ranks"][i]["player"]["name"])));
				}
				CurrentPage++;
			} catch {
				Error(ErrorType::Error, "RequestNextPage", "Cannot read rank values", getExceptionInfo() + "\n" + LeaderboardPageResponse);
				FailedRequestsCount++;
			}
			// Yield twice : Leaderboard will be drawn, scroll cursor will be higher so we dont insta-request the next page
			yield();yield();
			NewPageRequested = false;
		} else {
			Error(ErrorType::Warn, "RequestNextPage", "Invalid leaderboard page response!", LeaderboardPageResponse);
			FailedRequestsCount++;
			uint64 WaitingTime = Time::Now + uint(Math::Pow(2, FailedRequestsCount)) * 1000;
			while (Time::Now < WaitingTime) yield();
			NewPageRequested = false;
		}
	}
}

class LeaderboardPlayer {
	int Rank;
	int Points;
	Player@ Player;

	LeaderboardPlayer(int rank, int points, Player@ player) {
		Rank = rank;
		Points = points;
		@Player = player;
	}
}

class Player {
	string AccountID;
	string DisplayName;

	Player(const string &in accountID, const string &in displayName) {
		AccountID = accountID;
		DisplayName = displayName;
	}
}

class MatchParticipant {
	LeaderboardPlayer@ Player;
	uint TeamID;
	string PersonalBest = "...";
	uint DivisionPosition = 0;

	MatchParticipant(LeaderboardPlayer@ player, uint team) {
		@Player = player;
		TeamID = team;
	}
}

class Match {
	string Name;
	string LID = "";
	uint64 StartDate;
	uint64 EndDate;
	string JoinLink;
	string Status;
	string ServerStatus;
	array<Map@> Maps;
	string FirstMapName;
	array<MatchParticipant@> Players;

	array<uint> renderableTeams;
	uint teamNum = 0;

	uint QueueID = 0;

	void RefreshFromJSON(Json::Value JSON, bool ApplyMaps = false) {
		try {
			LID = JSON["liveId"];
			if (JSON["name"].GetType() != Json::Type::Null) Name = JSON["name"];
			if (JSON["startDate"].GetType() != Json::Type::Null) StartDate = JSON["startDate"];
			if (JSON["endDate"].GetType() != Json::Type::Null) EndDate = JSON["endDate"];
			if (JSON["joinLink"].GetType() != Json::Type::Null) JoinLink = JSON["joinLink"];
			if (JSON["status"].GetType() != Json::Type::Null) Status = JSON["status"];
			if (JSON["serverStatus"].GetType() != Json::Type::Null) ServerStatus = JSON["serverStatus"];
		} catch {
			Error(ErrorType::Error, "Match status", "Couldn't apply new values!", getExceptionInfo() + "\n" + Json::Write(JSON));
		}
	}

	void SetupPlayerData() {
		// require match state to be initialized and stuff
		if (LID == "") {
			Error(ErrorType::Error, "Match initial data", "Cannot setup initial match data when match is not initalized!", "");
			return;
		}

		QueueID = G_QueueState.CurrentQueue.QueueID;

		// Get players including their MM ranks from TM.io
		string MatchStatusResponse = SendJSONRequest(Net::HttpMethod::Get, "https://trackmania.io/api/match/" + LID);
		Json::Value MatchStatus = ResponseToJSON(MatchStatusResponse, Json::Type::Object);

		uint teams = 0;
		renderableTeams = array<uint>();
		if (MatchStatus.GetType() != Json::Type::Null && MatchStatus.HasKey("players")) {
			try {
				if (MatchStatus["players"].GetType() == Json::Type::Array) {
					// SendNadeoRequest for all players
					// For now, just record accountid and name
					// then gather points, rank
					// then gather PB
					Players.RemoveRange(0, Players.Length);
					for (uint i = 0; i < MatchStatus["players"].Length; i++) {
						int playerTeam = 0;
						try {
							playerTeam = MatchStatus["players"][i]["team"];
						} catch {
							Error(ErrorType::Log, "Match players info", "Team of a player couldnt be set!", getExceptionInfo());
						}
						try {
							Players.InsertLast(MatchParticipant(LeaderboardPlayer(C_Invalid, C_Invalid, Player(MatchStatus["players"][i]["player"]["id"], MatchStatus["players"][i]["player"]["name"])), MatchStatus["players"][i]["team"]));
							if (renderableTeams.Find(playerTeam) == -1) {
								renderableTeams.InsertLast(playerTeam);
							}
						} catch {
							Error(ErrorType::Log, "Match players info", "Player couldnt be added to list!", getExceptionInfo());
						}
						teams = Math::Max(teams, playerTeam);
					}
					startnew(FillParticipantPoints);
				}
			} catch {
				Error(ErrorType::Error, "Match players info", "Something went wrong while fetching the match participants!", getExceptionInfo() + "\n" + MatchStatusResponse);
			}
		} else {
			Error(ErrorType::Warn, "Trackmania.io match info", "Trackmania.io didnt know the players of the match :/", MatchStatusResponse);
		}

		if (MatchStatus.GetType() != Json::Type::Null && MatchStatus.HasKey("maps")) {
			try {
				if (MatchStatus["maps"].GetType() == Json::Type::Array) {
					if (MatchStatus["maps"].Length > 0) {
						FirstMapName = MatchStatus["maps"][0]["name"];
					}
				}
			} catch {
				Error(ErrorType::Error, "Match maps info", "Something went wrong while fetching the match maps!", getExceptionInfo() + "\n" + MatchStatusResponse);
			}
		} else {
			Error(ErrorType::Warn, "Trackmania.io match info", "Trackmania.io didnt know the maps of the match :/", MatchStatusResponse);
		}

		teamNum = teams + 1;
	}
}

void FillParticipantPoints() {
	string playerList = "";
	for (uint i = 0; i < G_QueueState.CurrentMatch.Players.Length; i++) {
		if (i != 0) playerList += "&";
		playerList += "players[]=" +  G_QueueState.CurrentMatch.Players[i].Player.Player.AccountID;
	}
	string ParticipantPointsResponse = SendNadeoRequest(Net::HttpMethod::Get, "https://meet.trackmania.nadeo.club/api/matchmaking/" + G_QueueState.CurrentQueue.QueueID + "/leaderboard/players?" + playerList);
	Json::Value ParticipantPoints = ResponseToJSON(ParticipantPointsResponse, Json::Type::Object);

	if (ParticipantPoints.GetType() != Json::Type::Null) {
		try {
			for (uint i = 0; i < ParticipantPoints["results"].Length; i++) {
				QueueRanking@ Ranking = QueueRanking();
				Ranking.Rank = ParticipantPoints["results"][i]["rank"];
				Ranking.Points = ParticipantPoints["results"][i]["score"];
				for (uint j = 0; j < G_QueueState.CurrentMatch.Players.Length; j++) {
					if (ParticipantPoints["results"][i]["player"] == G_QueueState.CurrentMatch.Players[j].Player.Player.AccountID) {
						G_QueueState.CurrentMatch.Players[j].Player.Rank = Ranking.Rank;
						G_QueueState.CurrentMatch.Players[j].Player.Points = Ranking.Points;
						if (G_QueueState.CurrentQueue.Stats.GetDivisionFromRanking(Ranking) !is null) G_QueueState.CurrentMatch.Players[j].DivisionPosition = G_QueueState.CurrentQueue.Stats.GetDivisionFromRanking(Ranking).Position;
						break;
					}
				}
			}
			// @LATER why was this removed lol
		} catch {
			Error(ErrorType::Error, "FillParticipantPoints", "Something went wrong while processing the Participant Points!", getExceptionInfo() + "\n" + ParticipantPointsResponse);
		}
	} else {
		Error(ErrorType::Warn, "FillParticipantPoints", "Leaderboard response is of Json type null!", ParticipantPoints);
	}
}

void FillParticipantPBs() {
	auto app = cast<CTrackMania>(GetApp());
	auto network = cast<CTrackManiaNetwork>(app.Network);
	// wait until score manager available
	while (network.ClientManiaAppPlayground is null || network.ClientManiaAppPlayground.ScoreMgr is null) yield();
	// request
	MwFastBuffer<wstring> RequestUserIds = MwFastBuffer<wstring>();
	for (uint i = 0; i < G_QueueState.CurrentMatch.Players.Length; i++) {
		RequestUserIds.Add(G_QueueState.CurrentMatch.Players[i].Player.Player.AccountID);
		G_QueueState.CurrentMatch.Players[i].PersonalBest = "???";
	}
	MwId userId;
	if (network.ClientManiaAppPlayground.UserMgr.Users.Length > 0) {
		userId = network.ClientManiaAppPlayground.UserMgr.Users[0].Id;
	} else {
		userId.Value = uint(-1);
	}
	CWebServicesTaskResult_MapRecordListScript@ recordList = network.ClientManiaAppPlayground.ScoreMgr.Map_GetPlayerListRecordList(userId, RequestUserIds, app.RootMap.MapInfo.MapUid, "PersonalBest", "", "TimeAttack", "");
	// analyze request
	while (recordList.IsProcessing) yield();
	if (recordList.HasSucceeded) {
		for (uint i = 0; i < recordList.MapRecordList.Length; i++) {
			for (uint j = 0; j < G_QueueState.CurrentMatch.Players.Length; j++) {
				if (recordList.MapRecordList[i].AccountId == G_QueueState.CurrentMatch.Players[j].Player.Player.AccountID || recordList.MapRecordList[i].WebServicesUserId == G_QueueState.CurrentMatch.Players[j].Player.Player.AccountID) {
					G_QueueState.CurrentMatch.Players[j].PersonalBest = Time::Format(recordList.MapRecordList[i].Time, true, false);
					break;
				}
			}
		}
	} else {
		Error(ErrorType::Log, "FillParticipantPBs_recordList", recordList.ErrorDescription, recordList.ErrorCode);
	}
}

class StatsMatch {
	string LID;
	int64 TimeStamp;
	array<Map@> Maps;
	int TeamResult;
	bool MVP;
	int PointsChange;
}

class Map {
	string FileLink;
	string Name;

	Map(const string &in url, const string &in name) {
		FileLink = url;
		Name = name;
	}
}

enum ErrorType {
	Warn,
	Error,
	Log,
	UserWarning,
	Info
}

// --- Plugin State

enum QueueStatus {
	NotInQueue,
	Pending,
	SkillGapCanceled,
	Queued,
	MatchFound,
	MatchAccepted,
	MatchReady,
	Canceled,
	CanceledUser,
	DifferentCodeCanceled,
	CrossplayCanceled,
	OldGameVersionCanceled,
	Banned,
	WaitForServer,
	Joining,
	InMatch,
	MatchCompleted,
	StartingQueue,
	Unknown
}

class PartyState {
	PluginSquad@ m_PluginSquad;

	Squad@ CurrentSquad;

	string PlayWith;
	uint64 LastSquadRefreshStamp;

	PartyState() {
		SquadMember@ MainPlayer = SquadMember(Player(GetMainUserInfo().WebServicesUserId, GetMainUserInfo().Name), true);

		@m_PluginSquad = PluginSquad();
		m_PluginSquad.Members.InsertLast(MainPlayer);
		@CurrentSquad = m_PluginSquad;
	}

	void RefreshPlayWith() {
		Json::Value Players = Json::Object();
		Json::Value PlayWithArray = Json::Array();
		if (CurrentSquad.Exists) {
			for (uint i = 0; i < CurrentSquad.Members.Length; i++) {
				if (CurrentSquad.Members[i].Player.AccountID != GetMainUserInfo().WebServicesUserId) PlayWithArray.Add(CurrentSquad.Members[i].Player.AccountID);
			}
		}
		Players["playWith"] = PlayWithArray;
		Players["code"] = G_DataState.PartyCode;
		PlayWith = Json::Write(Players);
	}
}

class Squad {
	array<SquadMember@> Members;
	array<Player@> InvitedPlayers;
	string Name;
	uint MaxSize = 3;

	bool Exists;

	void Leave() {
		Error(ErrorType::Error, "LeaveSquad", "Tried leaving a non-Nadeo Squad", "How did you get here????");
	}

	void Create() {
		Error(ErrorType::Error, "LeaveSquad", "Tried creating a non-Nadeo Squad", "How did you get here????");
	}

	void InvitePlayer(Player@ Player) {}
	void KickMember(Player@ player) {}

	void CancelInvitation(Player@ Player) {
		Error(ErrorType::Error, "CancelInvitation", "Tried canceling an invitation of a non-Nadeo Squad", "How did you get here????");
	}

	void AddMember(Player@ player) {
		if (Members.Length >= MaxSize) {
			Error(ErrorType::UserWarning, "AddMember", "Squad is already full!", "There would be too many members.");
			return;
		}
		Members.InsertLast(SquadMember(player));
		G_QueueState.Party.RefreshPlayWith();
	}

	void RemoveMember(Player@ player) {
		for (uint i = 0; i < Members.Length; i++) {
			if (player.AccountID == Members[i].Player.AccountID) {
				if (Members[i].Player.AccountID == GetMainUserInfo().WebServicesUserId) {
					Error(ErrorType::UserWarning, "KickMember", "You cannot remove yourself!", "Bruh");
					return;
				}
				if (Members[i].IsLeader) {
					Error(ErrorType::UserWarning, "KickMember", "You cannot remove the leader!", "If you are the leader yourself, leave instead!");
					return;
				}
				Members.RemoveAt(i);
				i--;
			}
		}
		G_QueueState.Party.RefreshPlayWith();
	}
}

class PluginSquad : Squad {

	PluginSquad() {
		Exists = true;
	}

	void InvitePlayer(Player@ Player) override {
		AddMember(Player);
	}

	void KickMember(Player@ Player) override {
		RemoveMember(Player);
	}
	
}

class SquadMember {
	Player@ Player;
	bool IsLeader;

	vec3 GraphColor = hsv2rgb(vec3(Math::Rand(0.0,1.0), 0.6, 1.0));
	float[] TMIOScoreProgression = {C_Invalid};

	SquadMember(Player@ player) {
		@Player = player;
		startnew(RefreshSquadMemberProgression, this);
	}

	SquadMember(Player@ player, bool isLeader) {
		@Player = player;
		startnew(RefreshSquadMemberProgression, this);
		IsLeader = isLeader;
	}
}

void RefreshSquadMemberProgression(ref@ a) {
	SquadMember@ Member = cast<SquadMember>(a);
	if (Member is null) {
		Error(ErrorType::Warn, "RefreshSquadMemberProgression", "Passed ref cast is null!", "");
		return;
	}

	Member.TMIOScoreProgression = GetPointsHistory(Member.Player.AccountID);
}

class UplayFriend {
	Player@ UplayPlayer;
	string Presence;
	string Relationship;
	string PlatformType;
	string WebServicesUserId;

	UplayFriend(CFriend@ Friend) {
		string AccountID = Friend.AccountId;
		string DisplayName = Friend.DisplayName;
		@UplayPlayer = Player(AccountID, DisplayName);
		Presence = Friend.Presence;
		Relationship = Friend.Relationship;
		PlatformType = Friend.PlatformType;
		WebServicesUserId = Friend.WebServicesUserId;
	}
}

class RecentTeammate {
	Player@ Player;
	int64 LastQueueStamp;
	
	RecentTeammate(Player@ player, int64 lastStamp) {
		@Player = player;
		LastQueueStamp = lastStamp;
	}
}

class QueueState {
	bool Ready = false;

	bool IsInQueue = false;
	QueueStatus Status;
	string StatusString;
	int BonkCounter;

	bool AutoRequeue = false;
	bool InputRequired = false;

	Match@ CurrentMatch;
	bool Absent;

	PartyState@ Party;

	array<Queue@> Queues;
	Queue@ CurrentQueue;

	uint64 LastHeartbeat;
	uint64 LastHeartbeatSent;
	uint64 QueueTimeStart;

	uint64 MatchFoundTime;

	void SetStatus(QueueStatus newStatus) {
		Status = newStatus;
		switch (Status) {
			case QueueStatus::NotInQueue :
				StatusString = "Not in Queue";
				break;
			case QueueStatus::Pending :
				StatusString = "Pending for partners";
				break;
			case QueueStatus::SkillGapCanceled :
				StatusString = "\\$f99Canceled (skill gap)";
				break;
			case QueueStatus::Queued :
				StatusString = "Queued";
				break;
			case QueueStatus::MatchFound :
				StatusString = "Match found";
				break;
			case QueueStatus::MatchAccepted :
				StatusString = "Waiting for players";
				break;
			case QueueStatus::MatchReady :
				StatusString = "Match ready!";
				break;
			case QueueStatus::Canceled :
				StatusString = "Canceled";
				break;
			case QueueStatus::CanceledUser :
				StatusString = "Canceled by user";
				break;
			case QueueStatus::DifferentCodeCanceled :
				StatusString = "\\$f99Canceled (check party code)";
				break;
			case QueueStatus::CrossplayCanceled :
				StatusString = "\\$f99Canceled (check crossplay settings)";
				break;
			case QueueStatus::OldGameVersionCanceled :
				StatusString = "\\$f99Canceled (old game version)";
				break;
			case QueueStatus::Banned :
				StatusString = "Banned";
				break;
			case QueueStatus::WaitForServer :
				StatusString = "Waiting for server";
				break;
			case QueueStatus::Joining :
				StatusString = "Joining";
				break;
			case QueueStatus::InMatch :
				StatusString = "In match";
				break;
			case QueueStatus::MatchCompleted :
				StatusString = "Match completed";
				break;
			case QueueStatus::StartingQueue :
				StatusString = "Starting queue...";
				break;
			default:
				StatusString = "Unknown";
		}
	}

	string GetStatusBarString() {

		// Party size
		string ReturnedValue = "\\$eee " + Icons::Users + " " + G_QueueState.Party.CurrentSquad.Members.Length;

		// Timer
		ReturnedValue += "  ";
		if (QueueTimeStart > 0) {
			// Time highscore color code
			if (Time::Now - QueueTimeStart > Setting_HighestWaitingTime) {
				ReturnedValue += CreateColorCode(hsv2rgb(vec3(0.175, Math::Abs(((Time::Now - QueueTimeStart) % 6000) - 3000) / 7500.0, 1.0)));
			} else {
				ReturnedValue += "\\$eee";
			}

			// Hourglass symbol
			int HourGlassValue = (Time::Stamp / 2) % 3;
			ReturnedValue += (HourGlassValue == 0 ? Icons::HourglassStart : (HourGlassValue == 1 ? Icons::HourglassHalf : Icons::HourglassEnd));

			// Timer
			ReturnedValue += Time::Format(Time::Now - QueueTimeStart, false);
		} else {
			// Display empty timer
			ReturnedValue += "\\$aaa" + Icons::Hourglass + Time::Format(0, false);
		}

		// Connection icon
		ReturnedValue += "  ";
		if (QueueTimeStart > 0) {
			if (Time::Now - Math::Max(LastHeartbeat,LastHeartbeatSent) < 10000) {
				ReturnedValue += CreateColorCode(hsv2rgb(vec3(0.33333 - Math::Abs(Time::Now - LastHeartbeat) / 30000.0, 0.8, 1.0)));
			} else {
				// No heartbeats in 10s. Sadge
				ReturnedValue += "\\$d00";
			}
		} else {
			ReturnedValue += "\\$aaa";
		}
		ReturnedValue += " " + (LastHeartbeat > LastHeartbeatSent ? Icons::Download : Icons::Upload);

		return ReturnedValue;
	}

	void AddQueue(Queue@ q) {
		if (Queues.Length == 0) @CurrentQueue = q;

		// Get and save divisions for the new queue

		string DivisionBordersResponse = SendNadeoRequest(Net::HttpMethod::Get, "https://meet.trackmania.nadeo.club/api/matchmaking/" + q.QueueID + "/division/display-rules");
		Json::Value DivisionBorders = ResponseToJSON(DivisionBordersResponse, Json::Type::Object);

		if (DivisionBorders.GetType() != Json::Type::Null && DivisionBorders.HasKey("divisions")) {
			try {
				q.Stats.Divisions.RemoveRange(0,q.Stats.Divisions.Length);
				for (uint i = 0; i < DivisionBorders["divisions"].Length; i++) {
					Division@ div = Division();
					div.DivisionID = DivisionBorders["divisions"][i]["id"];
					div.Position = DivisionBorders["divisions"][i]["position"];
					
					if (DivisionBorders["divisions"][i]["displayRuleType"] == "points_range") {
						div.RuleType = DivisionRuleType::PointsRange;
					} else if (DivisionBorders["divisions"][i]["displayRuleType"] == "victory_count" || DivisionBorders["divisions"][i]["displayRuleType"] == "minimum_victory_count") {
						div.RuleType = DivisionRuleType::VictoryCount;
					} else if (DivisionBorders["divisions"][i]["displayRuleType"] == "minimum_points") {
						div.RuleType = DivisionRuleType::MinimumPoints;
					} else if (DivisionBorders["divisions"][i]["displayRuleType"] == "minimum_rank") {
						div.RuleType = DivisionRuleType::MinimumRank;
					} else if (DivisionBorders["divisions"][i]["displayRuleType"] == "minimum_rank_and_points") {
						div.RuleType = DivisionRuleType::MinimumRankAndPoints;
					} else {
						q.Stats.CreateDefaultDivisions();
						Error(ErrorType::Log, "Parsing Division Borders", "Could not identify correct division rule type", tostring(DivisionBorders["divisions"][i]["displayRuleType"]) + " at " + DivisionBordersResponse);
						break;
					}
					
					if (DivisionBorders["divisions"][i]["displayRuleMinimumPoints"].GetType() == Json::Type::Number) div.MinPoints = DivisionBorders["divisions"][i]["displayRuleMinimumPoints"];
					if (DivisionBorders["divisions"][i]["displayRuleMaximumPoints"].GetType() == Json::Type::Number) div.MaxPoints = DivisionBorders["divisions"][i]["displayRuleMaximumPoints"];
					if (DivisionBorders["divisions"][i]["displayRuleMinimumRank"].GetType() == Json::Type::Number) div.MinRank = DivisionBorders["divisions"][i]["displayRuleMinimumRank"];
					if (DivisionBorders["divisions"][i]["displayRuleMinimumVictories"].GetType() == Json::Type::Number) div.MinVictories = DivisionBorders["divisions"][i]["displayRuleMinimumVictories"];
					if (DivisionBorders["divisions"][i]["displayRuleMaximumVictories"].GetType() == Json::Type::Number) div.MaxVictories = DivisionBorders["divisions"][i]["displayRuleMaximumVictories"];
					q.Stats.Divisions.InsertLast(div);
				}
			} catch {
				Error(ErrorType::Error, "AddQueue", "Something went wrong when parsing the division info response! Recreating default divisions...", getExceptionInfo() + "\n" + DivisionBordersResponse);
				q.Stats.CreateDefaultDivisions();
			}
		} else {
			Error(ErrorType::Error, "AddQueue", "Invalid division JSON received! Falling back to default.", DivisionBordersResponse);
		}

		// Division

		q.Stats.CurrentRanking = RefreshQueueRanking(q.QueueID);
		q.Stats.RefreshDivisionRanking();
		
		// @TODO
		// Now get the past matches n stuff by merging data from TM.io and the saved JSON
		// Progression, Matches
		RefreshPlayerStatus(q);
		// We know this queue exists, now get the current state
		
		// and finally, together with the stats, slot it in
		Queues.InsertLast(q);
	}

	void RefreshPlayerStatus(Queue@ q) {
		if (q.LastPlayerStatusCheck > Time::Stamp - 30) {
			Error(ErrorType::UserWarning, "RefreshPlayerStatus", "Too many requests!", "You can only refresh the maintenance status every 30 seconds. Please wait " + (30 + q.LastPlayerStatusCheck - Time::Stamp) + " more seconds!");
			return;
		}

		string PlayerStatusResponse = SendNadeoRequest(Net::HttpMethod::Get, "https://meet.trackmania.nadeo.club/api/matchmaking/" + q.QueueID + "/player-status");
		Json::Value PlayerStatus = ResponseToJSON(PlayerStatusResponse, Json::Type::Object);

		if (PlayerStatus.GetType() != Json::Type::Null) {
			try {
				if (PlayerStatus.HasKey("currentHeartbeat") && PlayerStatus["currentHeartbeat"].GetType() == Json::Type::Object) {
					if (GetStatusFromString(PlayerStatus["currentHeartbeat"]["status"]) == QueueStatus::Canceled) {
						Error(ErrorType::Log, "PlayerStatus", "Canceled player status. Not queueing.", PlayerStatusResponse);
					} else if (GetStatusFromString(PlayerStatus["currentHeartbeat"]["status"]) == QueueStatus::NotInQueue) {
						G_QueueState.SetStatus(GetStatusFromString(PlayerStatus["currentHeartbeat"]["status"]));
					} else {
						Error(ErrorType::Log, "RefreshPlayerStatus", "Starting new queue", PlayerStatus["currentHeartbeat"]["status"]);
						startnew(StartQueue, q);
					}
				}
				if (PlayerStatus.HasKey("matchmakingStatus") && PlayerStatus["matchmakingStatus"].GetType() == Json::Type::String) {
					q.Status = PlayerStatus["matchmakingStatus"];
				}
			} catch {
				Error(ErrorType::Warn, "AddQueue", "Something went wrong when getting the current heartbeat / penalty!", getExceptionInfo() + "\n" + PlayerStatusResponse);
			}
		} else {
			Error(ErrorType::Log, "AddQueue", "Invalid player status JSON received", PlayerStatusResponse);
		}

		q.LastPlayerStatusCheck = Time::Stamp;
	}
	
	void RefreshCurrentMatchStatus(const string &in LID, bool RefreshMaps = false) {
		string MatchStatusResponse = SendNadeoRequest(Net::HttpMethod::Get, "https://meet.trackmania.nadeo.club/api/matches/" + LID);
		Json::Value MatchStatusResult = ResponseToJSON(MatchStatusResponse, Json::Type::Object);
		if (MatchStatusResult.GetType() != Json::Type::Null && MatchStatusResult.HasKey("liveId")) {
			if (G_QueueState.CurrentMatch is null) {
				Error(ErrorType::Error, "Match info", "CurrentMatch is null!", MatchStatusResponse);
				return;
			}
			try {
				G_QueueState.CurrentMatch.RefreshFromJSON(MatchStatusResult, RefreshMaps);
			} catch {
				Error(ErrorType::Error, "Match status", "Couldn't apply refreshed match status!", getExceptionInfo() + "\n" + MatchStatusResponse);
			}
		} else {
			Error(ErrorType::Warn, "Match info", "Invalid response from server!", MatchStatusResponse);
		}
	}

	bool IncreaseBonkCounter() {
		BonkCounter++;

		if (BonkCounter > 1) {
			Error(ErrorType::UserWarning, "BonkCounter", "You failed to accept 2 matches!", "As a result, you have been stopped from queueing");
			ExitQueue();
			BonkCounter = 0;
			return true;
		} else {
			return false;
		}
	}

	QueueState() {
		@Party = PartyState();
		SetStatus(QueueStatus::NotInQueue);
	}
}

// UI System

class UIState {

	// UI values
	string LoadingString = "Loading...";
	bool InviteMenuVisible = false;
	bool ResizeMainWindow = true;
	bool ShowMatchLeaveUI = false;

	// UI data
	ColorScheme@ UserColor = ColorScheme();
	ColorScheme@ RankColor = ColorScheme();
	vec4 TabBarBottom = vec4(0.161,0.392,0.659, 1.0);

	array<Tab@> Tabs;
	array<Tab@> InviteTabs;

	// UI resources
	UI::Font@ Bold;
	dictionary MMTextures;
	dictionary RoyalTextures;
	dictionary RoyalPreviews;
	UI::Texture@ SuperRoyalTexture;

	// Sounds
	dictionary SoundSamples;

	// Colors
	ColorScheme@ Red = ColorScheme(vec4(0.647,0.176,0.157, 1.0), vec4(0.98,0.275,0.251, 1.0), vec4(0.98,0.027,0.059, 1.0));
	ColorScheme@ Green = ColorScheme(vec4(0.157,0.647,0.176, 1.0), vec4(0.251,0.98,0.275, 1.0), vec4(0.059,0.98,0.027, 1.0));
	ColorScheme@ Yellow = ColorScheme(vec4(0.85,0.69,0.08, 1.0), vec4(1.,0.812,0.129, 1.0), vec4(0.902,0.71,0.027, 1.0));
	ColorScheme@ Blue = ColorScheme();

	ColorScheme@ Gray = ColorScheme();

	UIState() {
		// Load tabs
		Tabs.InsertLast(QueueTab());
		Tabs.InsertLast(PartyTab());
		Tabs.InsertLast(StatsTab());
		Tabs.InsertLast(LeaderboardTab());
		Tabs.InsertLast(HistoryTab());
		Tabs.InsertLast(SettingsTab());

		// Load invite tabs
		InviteTabs.InsertLast(TMIOSearchTab());
		InviteTabs.InsertLast(FriendsTab());
		InviteTabs.InsertLast(RecentMembersTab());

		// Load resources
		@Bold = UI::LoadFont("DroidSans-Bold.ttf");
		for (uint i = 0; i <= 13; i++) {
			UI::Texture@ tex = UI::LoadTexture("Textures/3v3/" + i + ".png");
			if (tex is null) {
				Error(ErrorType::Error, "GetTexture", "Couldn't find the image 3v3/ " + i + ".png!", "");
			}
			MMTextures[tostring(i)] = @tex;
		}
		// its panic button hotfix time heya
		for (uint i = 0; i <= 5; i++) {
			UI::Texture@ tex = UI::LoadTexture("Textures/royal/" + i + ".png");
			if (tex is null) {
				Error(ErrorType::Error, "GetTexture", "Couldn't find the image royal/" + i + ".png!", "");
			}
			RoyalTextures[tostring(i)] = @tex;
		}
		for (uint i = 0; i <= 5; i++) {
			UI::Texture@ tex = UI::LoadTexture("Textures/royal/previews/" + i + ".png");
			if (tex is null) {
				Error(ErrorType::Error, "GetTexture", "Couldn't find the image royal/previews/" + i + ".png!", "");
			}
			RoyalPreviews[tostring(i)] = @tex;
		}
		@SuperRoyalTexture = UI::LoadTexture("Textures/super-royal/super-royal.png");
		if (SuperRoyalTexture is null) {
			Error(ErrorType::Error, "GetTexture", "Couldn't find the image super-royal/super-royal.png!", "");
		}
		Gray.SetColor(vec4(0.3,0.3,0.3,1.0));

		// Load sound samples
		SoundSamples["InvitationNotification"] = @Audio::LoadSample("Sounds/InvitationNotification.wav");
		SoundSamples["MatchFound"] = @Audio::LoadSample("Sounds/MatchFound.wav");
		SoundSamples["PlayButtonSelect"] = @Audio::LoadSample("Sounds/InvitationNotification.wav");
		SoundSamples["Race3"] = @Audio::LoadSample("Sounds/Race3.wav");
		SoundSamples["RaceWooshStart"] = @Audio::LoadSample("Sounds/RaceWooshStart.wav");
		SoundSamples["RankDown"] = @Audio::LoadSample("Sounds/RankDown.wav");
		SoundSamples["RankUp"] = @Audio::LoadSample("Sounds/RankUp.wav");
	}
}

class ColorScheme {

	// These are the default colors
	vec3 Color3;
	vec4 Color4;
	vec4 ColorHover;
	vec4 ColorActive;

	void SetColor(vec3 c) {
		SetColor(vec4(c.x,c.y,c.z,0.6));
	}

	void SetColor(vec4 c) {
		Color3 = vec3(c.x,c.y,c.z);
		Color4 = c;
		ColorHover = vec4(c.x + 0.075, c.y + 0.075, c.z + 0.075, c.w + 0.1);
		ColorActive = vec4(c.x + 0.16, c.y + 0.16, c.z + 0.16, c.w + 0.2);
	}

	ColorScheme() {
		// Set default colors
		Color3 = vec3(0.161,0.392,0.659);
		Color4 = vec4(0.161,0.392,0.659, 1.0);
		ColorHover = vec4(0.251,0.592,0.98, 1.0);
		ColorActive = vec4(0.027, 0.533, 0.98, 1.0);
	}

	ColorScheme(vec4 Normal, vec4 Hover, vec4 Active) {
		Color3 = vec3(Normal.x, Normal.y, Normal.z);
		Color4 = Normal;
		ColorHover = Hover;
		ColorActive = Active;
	}
}

class Tab {
	string Label;

	void Render() {
		UI::Text("you lost the game");
	}
}

class QueueTab : Tab {
	QueueTab() {
		Label = Icons::PlayCircle + " Queue";
	}
	
	void Render() override {
		// Player preview
		UI::BeginChild("AFKQueueTabPlayers", vec2(UI::GetWindowSize().x, UI::GetWindowSize().y - (28 * UI::GetScale())));

		if (G_QueueState.Status == QueueStatus::InMatch) {
			if (G_QueueState.Absent && G_UIState.ShowMatchLeaveUI) {
				RenderLeaveView();
			} else {
				RenderMatchView();
			}
		} else {
			RenderSquadView();
		}

		UI::EndChild();

		// Button bar
		RenderButtonBar();
	}

	void RenderMatchView() {
		RenderMatchTitle();
		RenderMatchParticipants();
	}

	void RenderLeaveView() {
		RenderMatchTitle();
		vec2 pos_orig = UI::GetCursorPos();

		string LeaveMatchString = "Leave Match?";
		vec2 LeaveMatchSize = Draw::MeasureString(LeaveMatchString, G_UIState.Bold);
		UI::SetCursorPos(vec2((UI::GetWindowSize().x - LeaveMatchSize.x) / 2, UI::GetCursorPos().y + 10));
		BoldText(LeaveMatchString);

		bool[] Buttons = {false, false};
		Buttons = RenderTwoButtons(Icons::SignOutAlt + " Leave", Icons::Hourglass + " Stay", G_UIState.Blue, G_UIState.Gray);
		if (Buttons[0]) {
			startnew(LeaveMatch);
		}
		if (Buttons[1]) {
			G_UIState.ShowMatchLeaveUI = false;
		}
	}

	void RenderMatchParticipants() {
		if (G_QueueState.CurrentMatch.teamNum == 0) return;

		string[] ColorCodes = {
			"\\$bcf",
			"\\$fcb"
		};

		UI::Columns(G_QueueState.CurrentMatch.renderableTeams.Length, "AFKQueueTabPlayersView", true);
		uint currentColumn = 0;
		UI::Dummy(vec2(0,0));
		UI::SetCursorPos(vec2(UI::GetCursorPos().x, UI::GetCursorPos().y + (6 * UI::GetScale())));
		float baseHeight = UI::GetCursorPos().y;
		int MVPPoints = GetMVPPoints();
		for (uint j = 0; j < G_QueueState.CurrentMatch.renderableTeams.Length; j++) { // @TODO 
			uint renderTeamID = G_QueueState.CurrentMatch.renderableTeams[j];
			
			for (uint i = 0; i < G_QueueState.CurrentMatch.Players.Length; i++) {
				if (G_QueueState.CurrentMatch.Players[i].TeamID != renderTeamID) continue;
				// Render Player
				// Div Name (Points)
				// Star Score (-Points)
				// <PB ICON> PB
				BoldText(
					GetDivisionTextShort(G_QueueState.CurrentMatch.Players[i].DivisionPosition, G_QueueState.CurrentMatch.QueueID) + " " +
					(G_QueueState.CurrentMatch.teamNum == 2 ? ColorCodes[G_QueueState.CurrentMatch.Players[i].TeamID] : "\\$fff") +
					G_QueueState.CurrentMatch.Players[i].Player.Player.DisplayName
				);UI::SameLine();
				RenderRight("\\$888" + G_QueueState.CurrentMatch.Players[i].Player.Points + "pts");
				BoldText(GetPlayerPoints(G_QueueState.CurrentMatch.Players[i].Player.Player.AccountID, MVPPoints)); UI::SameLine();
				UI::Text(GetMVPDelta(G_QueueState.CurrentMatch.Players[i].Player.Player.AccountID, MVPPoints)); UI::SameLine();
				
				RenderRight("\\$666" + GetPlacementFromRank(G_QueueState.CurrentMatch.Players[i].Player.Rank));
				UI::Separator();
			}
			UI::NextColumn();
			UI::Dummy(vec2(0,0));
			UI::SetCursorPos(vec2(UI::GetCursorPos().x, baseHeight));
		}
		UI::Columns(1);
	}

	int GetMVPPoints() {
		int maxPoints = C_Invalid;
		auto app = cast<CTrackMania>(GetApp());
		if (app.CurrentPlayground is null) return maxPoints;
		for (uint i = 0; i < app.CurrentPlayground.Players.Length; i++) {
			if (cast<CSmPlayer>(app.CurrentPlayground.Players[i]).Score !is null && cast<CSmPlayer>(app.CurrentPlayground.Players[i]).Score.Points > int(maxPoints)) {
				maxPoints = cast<CSmPlayer>(app.CurrentPlayground.Players[i]).Score.Points;
			}
		}
		return maxPoints;
	}

	string GetMVPDelta(const string &in AccountId, int MVPPoints) {
		auto app = cast<CTrackMania>(GetApp());
		if (app.CurrentPlayground is null) return "";
		for (uint i = 0; i < app.CurrentPlayground.Players.Length; i++) {
			if (app.CurrentPlayground.Players[i].User !is null && cast<CSmPlayer>(app.CurrentPlayground.Players[i]).Score !is null && app.CurrentPlayground.Players[i].User.WebServicesUserId == AccountId) {
				int delta = MVPPoints - cast<CSmPlayer>(app.CurrentPlayground.Players[i]).Score.Points;
				if (delta <= 0) return Icons::Star;
				return "\\$ccc(-" + tostring(delta) + ")";
			}
		}
		return "";
	}

	string GetPlayerPoints(const string &in AccountId, int MVPPoints) {
		// iterate through players, if player is found gib points
		auto app = cast<CTrackMania>(GetApp());
		if (app.CurrentPlayground is null) return "";
		for (uint i = 0; i < app.CurrentPlayground.Players.Length; i++) {
			if (app.CurrentPlayground.Players[i].User !is null && app.CurrentPlayground.Players[i].User.WebServicesUserId == AccountId) {
				if (cast<CSmPlayer>(app.CurrentPlayground.Players[i]).Score !is null) {
					return (cast<CSmPlayer>(app.CurrentPlayground.Players[i]).Score.Points == MVPPoints ? "\\$ffa" : "") + tostring(cast<CSmPlayer>(app.CurrentPlayground.Players[i]).Score.Points);
				} else return "";
			}
		}
		return "";
	}

	void RenderMatchTitle() {
		vec2 pos_orig = UI::GetCursorPos();

		string MatchTitle = (G_QueueState.CurrentQueue.QueueID == 2 && G_QueueState.CurrentMatch.FirstMapName.Length > 0 ? G_QueueState.CurrentMatch.FirstMapName : G_QueueState.CurrentMatch.Name) + " (" + G_QueueState.CurrentMatch.Status + ")";
		int Duration = Time::Stamp - G_QueueState.CurrentMatch.StartDate;
		string MatchDuration = (Duration > 1500 ? "\\$fcc" : "\\$cfc") + Time::Format(Duration * 1000, false);

		BoldText(MatchTitle); UI::SameLine();

		vec2 MatchDurationSize = Draw::MeasureString(MatchDuration);
		UI::SetCursorPos(vec2((UI::GetContentRegionMax().x - MatchDurationSize.x - 2), UI::GetCursorPos().y));
		UI::Text(MatchDuration);

		pos_orig = UI::GetCursorPos();
		UI::Separator();
		UI::SetCursorPos(pos_orig);
	}

	void RenderSquadView() {
		RenderSquadTitle();

		if (G_QueueState.Party.CurrentSquad.Members.Length < 1) return;

		UI::Columns(G_QueueState.Party.CurrentSquad.Members.Length, "AFKQueueTabPlayersView", false);
		for (uint i = 0; i < G_QueueState.Party.CurrentSquad.Members.Length; i++) {
			UI::BeginChild("AFKQueueTabPlayers" + i);

			RenderPlayerView(G_QueueState.Party.CurrentSquad.Members[i]);

			UI::EndChild();
			UI::NextColumn();
		}
		UI::Columns(1);
	}

	void RenderSquadTitle() {
		vec2 pos_orig = UI::GetCursorPos();

		vec2 SquadNameSize = Draw::MeasureString(G_QueueState.Party.CurrentSquad.Name, G_UIState.Bold);
		UI::PushStyleColor(UI::Col::Text, vec4(0.7,0.7,0.7,1.0));
		UI::SetCursorPos(vec2((UI::GetWindowSize().x - SquadNameSize.x) / 2, UI::GetWindowSize().y / 6 - 6));
		BoldText(G_QueueState.Party.CurrentSquad.Name);
		UI::PopStyleColor(1);

		UI::SetCursorPos(pos_orig);
	}

	void RenderPlayerView(SquadMember@ member) {
		vec2 pos_orig = UI::GetCursorPos();

		vec2 PlayerNameSize = Draw::MeasureString(member.Player.DisplayName, G_UIState.Bold);

		UI::SetCursorPos(vec2((-2 * UI::GetScale()),0));
		UI::PushStyleColor(UI::Col::FrameBg, vec4(0,0,0,0));
		UI::PushStyleColor(UI::Col::PlotLines, vec4(member.GraphColor.x, member.GraphColor.y, member.GraphColor.z ,0.2));
		UI::PushStyleColor(UI::Col::PlotLinesHovered, vec4(member.GraphColor.x, member.GraphColor.y, member.GraphColor.z ,0.3));
		UI::PushStyleColor(UI::Col::Text, vec4(0,0,0,0));
		UI::PushStyleColor(UI::Col::PopupBg, vec4(0,0,0,0));
		UI::SetNextItemWidth(UI::GetWindowSize().x + (4 * UI::GetScale()));
		if (member.TMIOScoreProgression[0] != C_Invalid) UI::PlotLines("##PlayerViewPoints", member.TMIOScoreProgression, 0, UI::GetWindowSize().y);
		UI::PopStyleColor(5);
		if (UI::IsItemHovered()) {
			vec4 graphRect = UI::GetItemRect();
			vec2 mousePos = UI::GetMousePos();
			mousePos = vec2(mousePos.x - graphRect.x, mousePos.y - graphRect.y);
			mousePos = vec2(mousePos.x / graphRect.z, mousePos.y / graphRect.w);
			uint index = uint(mousePos.x * member.TMIOScoreProgression.Length);
			UI::SetCursorPos(vec2(0,0));
			UI::PushStyleColor(UI::Col::Text, vec4(1,1,1,0.5));
			UI::Text(member.TMIOScoreProgression[index] + "pts");
			UI::PopStyleColor(1);
		}

		UI::SetCursorPos(vec2((UI::GetWindowSize().x - PlayerNameSize.x) / 2, UI::GetWindowSize().y / 2 - (14 * UI::GetScale())));
		BoldText(member.Player.DisplayName);

		UI::SetCursorPos(pos_orig);
	}

	void RenderButtonBar() {
		bool[] Buttons = {false, false};
		switch (G_QueueState.Status) {
			case QueueStatus::SkillGapCanceled :
			case QueueStatus::NotInQueue :
			case QueueStatus::MatchCompleted :
			case QueueStatus::Banned :
			case QueueStatus::CanceledUser :
			case QueueStatus::DifferentCodeCanceled :
			case QueueStatus::CrossplayCanceled :
			case QueueStatus::OldGameVersionCanceled :
				RenderPartyCode();
				if (RenderOneButton(Icons::Play + " Enter Queue", (Setting_ColoredUI ? G_UIState.UserColor : null)) && !G_QueueState.IsInQueue) {
					if (G_QueueState.CurrentQueue.Status == "under_maintenance") {
						Error(ErrorType::Info, "Starting Queue Button", "This queue is under maintenance!", "Starting a maintenance status check...");
						startnew(PlayerStatusCheck, G_QueueState.CurrentQueue);
						break;
					}
					G_QueueState.IsInQueue = true;
					startnew(QueueLoop);
				}
				break;
			case QueueStatus::Pending :
			case QueueStatus::Queued :
			case QueueStatus::Canceled :
			case QueueStatus::StartingQueue :
				RenderPartyCode();
				if (RenderOneButton(Icons::Stop + " Exit Queue", (Setting_ColoredUI ? G_UIState.Red : null)) && G_QueueState.IsInQueue) {
					G_QueueState.IsInQueue = false;
					ExitQueue();
				}
				break;
			case QueueStatus::MatchFound :
				Setting_AutoAccept = RenderSettingChoice(Setting_AutoAccept, "Auto-Accept");
				if (!Setting_AutoAccept) {
					Buttons = RenderTwoButtons(Icons::Check + " Accept", Icons::Times + " Cancel", G_UIState.Green, G_UIState.Red);
					if (Buttons[0]) {
						startnew(SendAccept);
						G_QueueState.SetStatus(QueueStatus::MatchAccepted);
					}
					if (Buttons[1]) {
						G_QueueState.IsInQueue = false;
						ExitQueue();
					}
				} else if (G_QueueState.InputRequired) {
					if (RenderOneButton("PLEASE PROVIDE INPUT", G_UIState.Green)) {
						startnew(SendAccept);
						startnew(PlayMatchFoundInputReceivedSound);
						G_QueueState.InputRequired = false;
						G_QueueState.SetStatus(QueueStatus::MatchAccepted);
					}
				}
				break;
			case QueueStatus::MatchAccepted :
			case QueueStatus::MatchReady :
				break;
			case QueueStatus::WaitForServer :
				Buttons = RenderTwoButtons("Pinging Server...", Icons::Clipboard, G_UIState.Gray, G_UIState.Yellow);
				if (Buttons[1]) {
					CopyMatchID();
				}
				break;
			case QueueStatus::Joining :
				Buttons = RenderTwoButtons(Icons::Play + " Join manually", Icons::Clipboard, G_UIState.Blue, G_UIState.Yellow);
				if (Buttons[0]) {
					JoinServer();
				}
				if (Buttons[1]) {
					CopyMatchID();
				}
				break;
			case QueueStatus::InMatch :
				G_QueueState.AutoRequeue = RenderSettingChoice(G_QueueState.AutoRequeue, "Auto-Requeue");
				Buttons = RenderTwoButtons(Icons::Clipboard + " Copy MatchID", Icons::Play, G_UIState.Yellow, G_UIState.Blue);
				if (Buttons[0]) {
					CopyMatchID();
				}
				if (Buttons[1]) {
					JoinServer();
				}
				break;
		}
	}

	// - Layout: Center label
	void RenderLabel(const string &in Label) {
		vec2 pos_orig = UI::GetCursorPos();

		vec2 LabelSize = Draw::MeasureString(Label, G_UIState.Bold);
		UI::SetCursorPos(vec2((UI::GetWindowSize().x - LabelSize.x) / 2, UI::GetCursorPos().y));
		BoldText(Label);

		UI::SetCursorPos(pos_orig);
	}

	// - Layout: Right label
	void RenderRight(const string &in Label) {
		vec2 pos_orig = UI::GetCursorPos();
		vec2 LabelSize = Draw::MeasureString(Label);
		UI::SetCursorPos(vec2((UI::GetContentRegionMax().x - LabelSize.x - (2 * UI::GetScale())), UI::GetCursorPos().y));
		UI::Text(Label);
	}

	// - Layout: One main button
	bool RenderOneButton(const string &in Label, ColorScheme@ Color) {
		vec2 pos_orig = UI::GetCursorPos();
		bool ReturnedValue;

		vec2 ButtonSize = Draw::MeasureString(Label + "__");

		if (Color !is null) {
			UI::PushStyleColor(UI::Col::Button, Color.Color4);
			UI::PushStyleColor(UI::Col::ButtonHovered, Color.ColorHover);
			UI::PushStyleColor(UI::Col::ButtonActive, Color.ColorActive);
		}

		UI::SetCursorPos(vec2((UI::GetWindowSize().x - ButtonSize.x) / 2, UI::GetWindowSize().y - (24 * UI::GetScale())));
		ReturnedValue = UI::Button(Label);

		if (Color !is null) UI::PopStyleColor(3);

		UI::SetCursorPos(pos_orig);
		return ReturnedValue;
	}

	// - Layout: Two Buttons
	bool[] RenderTwoButtons(const string &in LabelA, const string &in LabelB, ColorScheme@ ColorA, ColorScheme@ ColorB) {
		vec2 pos_orig = UI::GetCursorPos();
		bool[] ReturnedValue = {false,false};

		vec2 ButtonASize = Draw::MeasureString(LabelA + "__");
		vec2 TotalSize = Draw::MeasureString(LabelA + "__" + LabelB + "__");

		if (ColorA !is null) {
			UI::PushStyleColor(UI::Col::Button, ColorA.Color4);
			UI::PushStyleColor(UI::Col::ButtonHovered, ColorA.ColorHover);
			UI::PushStyleColor(UI::Col::ButtonActive, ColorA.ColorActive);
		}

		UI::SetCursorPos(vec2((UI::GetWindowSize().x - TotalSize.x) / 2 - (2 * UI::GetScale()), UI::GetWindowSize().y - (24 * UI::GetScale())));
		ReturnedValue[0] = UI::Button(LabelA);

		if (ColorA !is null) UI::PopStyleColor(3);

		if (ColorB !is null) {
			UI::PushStyleColor(UI::Col::Button, ColorB.Color4);
			UI::PushStyleColor(UI::Col::ButtonHovered, ColorB.ColorHover);
			UI::PushStyleColor(UI::Col::ButtonActive, ColorB.ColorActive);
		}

		UI::SetCursorPos(vec2((UI::GetWindowSize().x - TotalSize.x) / 2 + ButtonASize.x + (4 * UI::GetScale()), UI::GetWindowSize().y - (24 * UI::GetScale())));
		ReturnedValue[1] = UI::Button(LabelB);

		if (ColorB !is null) UI::PopStyleColor(3);

		UI::SetCursorPos(pos_orig);
		return ReturnedValue;
	}

	// - Setting checkbox
	bool RenderSettingChoice(bool Value, const string &in Label) {
		bool newValue;
		vec2 pos_orig = UI::GetCursorPos();
		UI::SetCursorPos(vec2(0, UI::GetWindowSize().y - (24 * UI::GetScale())));
		newValue = UI::Checkbox(Label, Value);
		UI::SetCursorPos(pos_orig);
		return newValue;
	}

	void RenderPartyCode() {
		vec2 pos_orig = UI::GetCursorPos();
		if (G_QueueState.IsInQueue) {
			UI::Text(G_DataState.PartyCode);
		} else {
			UI::SetCursorPos(vec2(pos_orig.x, pos_orig.y - (2 * UI::GetScale())));
			UI::SetNextItemWidth(82);
			G_DataState.PartyCode = UI::InputText("Party Code", G_DataState.PartyCode);
			UI::SetCursorPos(vec2(UI::GetWindowSize().x - Draw::MeasureString("_______Code").x, pos_orig.y - (2 * UI::GetScale())));
			if (UI::Button(Icons::PlusSquare + " Code")) {
				startnew(GenerateNewPartyCode);
			}
		}
		UI::SetCursorPos(pos_orig);
	}

}

class PartyTab : Tab {
	PartyTab() {
		Label = Icons::UserCircle + " Party";
	}

	void Render() override {
		if (UI::Button(Icons::Users + "  Add Players")) {
			startnew(GetFriendList);
			G_UIState.InviteMenuVisible = true;
		}
		UI::Text("Reminder: Other players will need to add you aswell!");
		RenderSquadMembers();
	}

	void RenderSquadMembers() {
		UI::BeginChild("AFKSquadMembers");
		if (UI::BeginTable("SquadMembers", 2, UI::TableFlags::ScrollY)) {
			UI::TableSetupScrollFreeze(0, 1);
			UI::TableSetupColumn(" ", UI::TableColumnFlags::WidthFixed, (30 * UI::GetScale()));
			UI::TableSetupColumn("Members", UI::TableColumnFlags::WidthStretch);
			UI::TableHeadersRow();

			bool MainUserIsLeader = false;
			for (uint i = 0; i < G_QueueState.Party.CurrentSquad.Members.Length; i++) {
				if (G_QueueState.Party.CurrentSquad.Members[i].Player.AccountID == GetMainUserInfo().WebServicesUserId) {
					MainUserIsLeader = G_QueueState.Party.CurrentSquad.Members[i].IsLeader;
					break;
				}
			}

			for (uint i = 0; i < G_QueueState.Party.CurrentSquad.Members.Length; i++) {
				UI::PushID("SquadMember" + i);
				if (G_QueueState.Party.CurrentSquad.Members[i].Player.AccountID != GetMainUserInfo().WebServicesUserId) {
					UI::TableNextRow();
					
					UI::TableSetColumnIndex(0);
					UI::PushStyleColor(UI::Col::Button, G_UIState.Red.Color4);
					UI::PushStyleColor(UI::Col::ButtonHovered, G_UIState.Red.ColorHover);
					UI::PushStyleColor(UI::Col::ButtonActive, G_UIState.Red.ColorActive);
					if (!G_QueueState.Party.CurrentSquad.Members[i].IsLeader && MainUserIsLeader && UI::Button(Icons::Times)) {
						G_QueueState.Party.CurrentSquad.KickMember(G_QueueState.Party.CurrentSquad.Members[i].Player);
						UI::PopStyleColor(3);
						UI::PopID();
						break;
					}
					UI::PopStyleColor(3);

					UI::TableSetColumnIndex(1);
					UI::Text(G_QueueState.Party.CurrentSquad.Members[i].Player.DisplayName);
				} else {
					UI::TableNextRow();
					UI::TableSetColumnIndex(1);
					UI::Text("\\$999" + G_QueueState.Party.CurrentSquad.Members[i].Player.DisplayName);
				}
				UI::PopID();
			}

			for (uint i = 0; i < G_QueueState.Party.CurrentSquad.InvitedPlayers.Length; i++) {
				UI::PushID("InvitedMember" + i);
				UI::TableNextRow();

				UI::TableSetColumnIndex(0);
				UI::PushStyleColor(UI::Col::Button, G_UIState.Red.Color4);
				UI::PushStyleColor(UI::Col::ButtonHovered, G_UIState.Red.ColorHover);
				UI::PushStyleColor(UI::Col::ButtonActive, G_UIState.Red.ColorActive);
				if (UI::Button(Icons::UserTimes)) {
					G_QueueState.Party.CurrentSquad.CancelInvitation(G_QueueState.Party.CurrentSquad.InvitedPlayers[i]);
					UI::PopStyleColor(3);
					UI::PopID();
					break;
				}
				UI::PopStyleColor(3);

				UI::TableSetColumnIndex(1);
				UI::Text(G_QueueState.Party.CurrentSquad.InvitedPlayers[i].DisplayName + " \\$999(Invited)");
				UI::PopID();
			}

			UI::EndTable();
		}
		UI::EndChild();
	}
}

class StatsTab : Tab {
	StatsTab() {
		Label = Icons::ChartLine + " Stats";
	}

	void Render() override {
		// I have decided that Trackmania.io stats tracking is better because it saves the point progression more accurately.
		// If I get back to this, you will be able to customize the timeframe of past matches as tracked by Trackmania.io
		
		RenderLabel("View your stats on Trackmania.io:");
		if (RenderOneButton(Icons::ExternalLinkAlt + "  View", (Setting_ColoredUI ? G_UIState.UserColor : null))) {
			OpenBrowserURL("https://trackmania.io/#/player/" + GetMainUserInfo().WebServicesUserId);
		}
	}

	void RenderLabel(const string &in Label) {
		vec2 pos_orig = UI::GetCursorPos();

		vec2 LabelSize = Draw::MeasureString(Label, G_UIState.Bold);
		UI::SetCursorPos(vec2((UI::GetWindowSize().x - LabelSize.x) / 2, UI::GetWindowSize().y / 2 - 14));
		BoldText(Label);

		UI::SetCursorPos(pos_orig);
	}

	// - Layout: One main button
	bool RenderOneButton(const string &in Label, ColorScheme@ Color) {
		vec2 pos_orig = UI::GetCursorPos();
		bool ReturnedValue;

		vec2 ButtonSize = Draw::MeasureString(Label + "__");

		if (Color !is null) {
			UI::PushStyleColor(UI::Col::Button, Color.Color4);
			UI::PushStyleColor(UI::Col::ButtonHovered, Color.ColorHover);
			UI::PushStyleColor(UI::Col::ButtonActive, Color.ColorActive);
		}

		UI::SetCursorPos(vec2((UI::GetWindowSize().x - ButtonSize.x) / 2, UI::GetWindowSize().y / 2 + (14 * UI::GetScale())));
		ReturnedValue = UI::Button(Label);

		if (Color !is null) UI::PopStyleColor(3);

		UI::SetCursorPos(pos_orig);
		return ReturnedValue;
	}
}

class LeaderboardTab : Tab {
	LeaderboardTab() {
		Label = Icons::Trophy + " Top";
	}

	void Render() override {
		if (UI::BeginTable("LeaderboardTop", 3, UI::TableFlags::ScrollY)) {
			UI::TableSetupScrollFreeze(0, 1);
			UI::TableSetupColumn("Rank", UI::TableColumnFlags::WidthFixed, (32 * UI::GetScale()));
			UI::TableSetupColumn("Points", UI::TableColumnFlags::WidthFixed, (40 * UI::GetScale()));
			UI::TableSetupColumn("Player", UI::TableColumnFlags::WidthStretch);
			
			UI::TableHeadersRow();

			vec2 pos_orig = UI::GetCursorPos();
			UI::PushStyleColor(UI::Col::Button, vec4(G_UIState.Blue.Color3.x, G_UIState.Blue.	Color3.y, G_UIState.Blue.Color3.z, 0));
			UI::PushStyleColor(UI::Col::ButtonHovered, vec4(G_UIState.Blue.Color3.x, 	G_UIState.Blue.Color3.y, G_UIState.Blue.Color3.z, 0.4));
			UI::PushStyleColor(UI::Col::ButtonActive, vec4(G_UIState.Blue.Color3.x, G_UIState.	Blue.Color3.y, G_UIState.Blue.Color3.z, 0.7));
			UI::SetCursorPos(vec2(UI::GetWindowSize().x - (48 * UI::GetScale()), pos_orig.y - (21 * UI::GetScale())));
			if (UI::Button(Icons::Redo)) {
				if (G_QueueState.CurrentQueue.Stats.Leaderboard.LastRequestStamp + 300000 < 	Time::Now) {
					G_QueueState.CurrentQueue.Stats.Leaderboard.CurrentPage = 0;
					G_QueueState.CurrentQueue.Stats.Leaderboard.PlayerLeaderboard = {};
				} else {
					Error(ErrorType::Info, "RefreshButton", "You can't refresh yet!", "Please 	wait " + Math::Floor((G_QueueState.CurrentQueue.Stats.Leaderboard.	LastRequestStamp+ 300000 - Time::Now) / 1000) + " more seconds until you 	can refresh. The Trackmania.io Leaderboard refreshes every few minutes 	and is not real-time.");
				}
			}
			UI::PopStyleColor(3);
			UI::SetCursorPos(pos_orig);
			UI::Dummy(vec2(0,0));

			UI::ListClipper Clipper(G_QueueState.CurrentQueue.Stats.Leaderboard.PlayerLeaderboard.Length);

			while (Clipper.Step()) {
				for (int i = Clipper.DisplayStart; i < Clipper.DisplayEnd; i++) {
					UI::TableNextRow();

					UI::TableSetColumnIndex(0);
					UI::Text(tostring(G_QueueState.CurrentQueue.Stats.Leaderboard.PlayerLeaderboard[i].Rank));
					UI::TableSetColumnIndex(1);
					UI::Text(tostring(G_QueueState.CurrentQueue.Stats.Leaderboard.PlayerLeaderboard[i].Points));
					UI::TableSetColumnIndex(2);
					UI::Text(tostring(G_QueueState.CurrentQueue.Stats.Leaderboard.PlayerLeaderboard[i].Player.DisplayName));
				}
			}

			if (G_QueueState.CurrentQueue.Stats.Leaderboard.LastPageEmpty) {
				UI::TableNextRow();
				UI::TableSetColumnIndex(2);
				UI::Text("End of Leaderboard (last page was empty)");
			} else if (!G_QueueState.CurrentQueue.Stats.Leaderboard.NewPageRequested && (UI::GetScrollY() >= UI::GetScrollMaxY() - 24)) {
				// Start request
				G_QueueState.CurrentQueue.Stats.Leaderboard.NewPageRequested = true;
				startnew(RequestNextLeaderboardPage);
			} else if (G_QueueState.CurrentQueue.Stats.Leaderboard.NewPageRequested) {
				UI::TableNextRow();
				UI::TableSetColumnIndex(0);
				UI::Text("");
				UI::TableSetColumnIndex(1);
				UI::Text("");
				UI::TableSetColumnIndex(2);
				UI::Text("Loading...");
			}

			UI::EndTable();
		}
		
	}
}

class HistoryTab : Tab {
	HistoryTab() {
		Label = Icons::Book + " History";
	}

	void Render() override {

		if (UI::BeginTable("MatchHistory", 4, UI::TableFlags::ScrollY)) {
			UI::TableSetupScrollFreeze(0, 1);
			UI::TableSetupColumn("Result", UI::TableColumnFlags::WidthFixed, (64 * UI::GetScale()));
			UI::TableSetupColumn("Date", UI::TableColumnFlags::WidthFixed, UI::GetWindowSize().x * 0.3);
			UI::TableSetupColumn("Map", UI::TableColumnFlags::WidthFixed, UI::GetWindowSize().x * 0.3);
			UI::TableSetupColumn("Action", UI::TableColumnFlags::WidthStretch);
			UI::TableHeadersRow();

			UI::ListClipper Clipper(G_QueueState.CurrentQueue.Stats.PlayerMatches.Length);

			while (Clipper.Step()) {
				for (int i = Clipper.DisplayStart; i < Clipper.DisplayEnd; i++) {
					UI::TableNextRow();
					UI::PushID("MatchHistory" + i);

					UI::TableSetColumnIndex(0);
					if (G_QueueState.CurrentQueue.Stats.PlayerMatches[i].TeamResult == 1) {
						UI::Text("\\$afa" + Icons::Trophy);
					} else {
						BoldText("\\$ccc" + GetPlacementFromRank(G_QueueState.CurrentQueue.Stats.PlayerMatches[i].TeamResult));
					}
					if (G_QueueState.CurrentQueue.Stats.PlayerMatches[i].MVP) {
						UI::SameLine();
						UI::Text(Icons::Star);
					}
					UI::TableSetColumnIndex(1);
					UI::Text(Time::FormatString("%x %X",G_QueueState.CurrentQueue.Stats.PlayerMatches[i].TimeStamp));
					UI::TableSetColumnIndex(2);
					string MapString = "";
					if (G_QueueState.CurrentQueue.Stats.PlayerMatches[i].Maps.Length < 1) {
						MapString = "???";
					} else {
						MapString = G_QueueState.CurrentQueue.Stats.PlayerMatches[i].Maps[0].Name;
						if (G_QueueState.CurrentQueue.Stats.PlayerMatches[i].Maps.Length > 1) MapString += ", \\$888... (" + (G_QueueState.CurrentQueue.Stats.PlayerMatches[i].Maps.Length - 1) + ")";
					}
					UI::Text(MapString);
					UI::TableSetColumnIndex(3);
					if (UI::Button(Icons::ExternalLinkAlt + " TM.io")) {
						OpenBrowserURL("https://trackmania.io/#/match/" + G_QueueState.CurrentQueue.Stats.PlayerMatches[i].LID);
					}
					UI::PopID();
				}
			}
			
			UI::EndTable();
		}
	}
}

class SettingsTab : Tab {
	SettingsTab() {
		Label = Icons::Wrench + " Settings";
	}

	void Render() override {
		UI::PushStyleColor(UI::Col::Separator, vec4(0.3, 0.3, 0.3, 0.6));

		BoldText("Sounds");
		Setting_SoundVolume = UI::SliderFloat("Notification Volume", Setting_SoundVolume, 0, 100);
		if (UI::Button(Icons::Bell + " Test sound")) startnew(PlayInfoSound);

		UI::Separator();

		BoldText("UI");
		Setting_ColoredUI = UI::Checkbox("Use custom colors", Setting_ColoredUI);
		if (UI::IsItemHovered()) {
			UI::BeginTooltip();
			UI::Text("Enable this to have the menu tabs colored after your Matchmaking rank!");
			UI::Text("The \"Enter Queue\" button is also colored after your Trails color in the TM Profile settings.");
			UI::EndTooltip();
		}
		
		UI::Separator();

		BoldText("Notifications");
		UI::SetCursorPos(vec2(UI::GetCursorPos().x, UI::GetCursorPos().y + (2 * UI::GetScale())));
		UI::Text("Play invitiation sounds for: ");
		UI::SameLine();
		UI::SetCursorPos(vec2(UI::GetCursorPos().x, UI::GetCursorPos().y - (4  * UI::GetScale())));
		UI::SetNextItemWidth((UI::GetWindowSize().x - UI::GetCursorPos().x) / UI::GetScale() - 32);
		if (UI::BeginCombo("##InvitationNotificationSettingCombo", G_DataState.NotificationSettings[Setting_ShowInviteNotif])) {
			// probably not the best way to get rid of an unsigned / signed mismatch, but I cant have a uint type setting xd
			for (int i = 0; uint(i) < G_DataState.NotificationSettings.Length; i++) {
				if (UI::Selectable(G_DataState.NotificationSettings[i], i == Setting_ShowInviteNotif)) {
					Setting_ShowInviteNotif = i;
				}
			}
			UI::EndCombo();
		}

		Setting_HideWarn = UI::Checkbox("Hide silent warning popups", Setting_HideWarn);
		if (UI::IsItemHovered()) {
			UI::BeginTooltip();
			UI::Text("These popups should only concern you if they repeatedly happen every ~5s or so.");
			UI::Text("You can mute them if you want (but don't complain about stuff stopping to work without you noticing!)");
			UI::EndTooltip();
		}

		// @TODO register to the financial govt as someone that is able to take donations online. Yes, it's that complicated legally
		// Hey. If you read this, hope you're having fun with the plugin!
		// While I do all my plugin stuff and work in my free time, a lot of effort went into this and
		// I have been almost overworking myself at some point.
		// I was asked recently why I wouldn't pay some TM twitch subs I got gifted myself
		// - Well, I don't feel like I got the money to spend it on twitch subs for example.
		// Everthing I do here has been and is non-profit and you should never have to pay me to use plugins like this.
		// Already the pressure of people "wanting to get something in return for the money" and potentially jump into burn-out cause of that
		// keeps me from doing it. But I have been thinking lately of a way to support what I do, fully optional, in case you like what I do.
		// I'd wholeheartedly appreciate any support, let it be a paypal donation, a giftsub on twitch or whatever.
		// It motivates me to do more awesome stuff like this. Thank you.
		/*
		UI::Separator();
		BoldText("Support Me!");
		if (UI::Button(Icons::Brands::Paypal + " Paypal Donation (fully optional)")) {
			//<paypal.me link or smth idk>
		}
		*/

		UI::PopStyleColor(1);
	}
}

class TMIOSearchTab : Tab {
	TMIOSearchTab() {
		Label = Icons::Search + " TM.io Search";
	}

	void Render() override {
		// Search bar
		bool StartSearch = false;
		G_DataState.TMIOSearchString = UI::InputText("Search", G_DataState.TMIOSearchString, StartSearch, UI::InputTextFlags::EnterReturnsTrue | (G_DataState.TMIOSearchResultsValid ? 0 : UI::InputTextFlags::ReadOnly));
		if (StartSearch) startnew(TMIOSearch);

		// Search results

		if (!G_DataState.TMIOSearchResultsValid) {
			string SearchingString = "Searching...";
			vec2 SearchingSize = Draw::MeasureString(SearchingString);
			UI::SetCursorPos((UI::GetWindowSize() - SearchingSize) / 2);
			UI::Text(SearchingString);
			return;
		}
		
		if (UI::BeginTable("TMIOSearchResults", 3, UI::TableFlags::ScrollY)) {
			UI::TableSetupScrollFreeze(0,1);
			UI::TableSetupColumn(" ", UI::TableColumnFlags::WidthFixed, (30 * UI::GetScale()));
			UI::TableSetupColumn("Name", UI::TableColumnFlags::WidthStretch);
			UI::TableSetupColumn("Points", UI::TableColumnFlags::WidthFixed, UI::GetWindowSize().x * 0.3);
			UI::TableHeadersRow();

			UI::ListClipper Clipper(G_DataState.TMIOSearchResults.Length);
			while (Clipper.Step()) {

				for (int i = Clipper.DisplayStart; i < Clipper.DisplayEnd; i++) {
					UI::TableNextRow();
					UI::PushID("TMIOUser" + i);

					// Button Column
					UI::TableSetColumnIndex(0);
					bool UserInParty = false;
					bool UserLocked = false;
					bool UserInvited = false;

					for (uint j = 0; j < G_QueueState.Party.CurrentSquad.Members.Length; j++) {
						if (G_DataState.TMIOSearchResults[i].Player.AccountID == G_QueueState.Party.CurrentSquad.Members[j].Player.AccountID) {
							UserInParty = true;
							if (G_QueueState.Party.CurrentSquad.Members[j].IsLeader || G_QueueState.Party.CurrentSquad.Members[j].Player.AccountID == GetMainUserInfo().WebServicesUserId) UserLocked = true;
							break;
						}
					}

					for (uint j = 0; j < G_QueueState.Party.CurrentSquad.InvitedPlayers.Length; j++) {
						if (G_DataState.TMIOSearchResults[i].Player.AccountID == G_QueueState.Party.CurrentSquad.InvitedPlayers[j].AccountID) {
							UserInvited = true;
							break;
						}
					}

					// Decide button color
					if (UserLocked) {
						UI::PushStyleColor(UI::Col::Button, G_UIState.Gray.Color4);
						UI::PushStyleColor(UI::Col::ButtonHovered, G_UIState.Gray.ColorHover);
						UI::PushStyleColor(UI::Col::ButtonActive, G_UIState.Gray.ColorActive);
					} else if (UserInParty) {
						UI::PushStyleColor(UI::Col::Button, G_UIState.Red.Color4);
						UI::PushStyleColor(UI::Col::ButtonHovered, G_UIState.Red.ColorHover);
						UI::PushStyleColor(UI::Col::ButtonActive, G_UIState.Red.ColorActive);
					} else if (UserInvited) {
						UI::PushStyleColor(UI::Col::Button, G_UIState.Blue.Color4);
						UI::PushStyleColor(UI::Col::ButtonHovered, G_UIState.Blue.ColorHover);
						UI::PushStyleColor(UI::Col::ButtonActive, G_UIState.Blue.ColorActive);
					} else {
						UI::PushStyleColor(UI::Col::Button, G_UIState.Green.Color4);
						UI::PushStyleColor(UI::Col::ButtonHovered, G_UIState.Green.ColorHover);
						UI::PushStyleColor(UI::Col::ButtonActive, G_UIState.Green.ColorActive);
					}

					// Decide button label / action
					if (UserInParty) {
						if (UI::Button(Icons::UserTimes)) {
							G_QueueState.Party.CurrentSquad.KickMember(G_DataState.TMIOSearchResults[i].Player);
						}
					} else {
						if (UI::Button(Icons::UserPlus) && !UserInvited) {
							G_QueueState.Party.CurrentSquad.InvitePlayer(G_DataState.TMIOSearchResults[i].Player);
						}
					}

					UI::PopStyleColor(3);

					// Name
					UI::TableSetColumnIndex(1);
					UI::Text(G_DataState.TMIOSearchResults[i].Player.DisplayName);

					// Ranking
					UI::TableSetColumnIndex(2);
					if (G_DataState.TMIOSearchResults[i].Points != C_Invalid || G_DataState.TMIOSearchResults[i].Rank != C_Invalid)
						UI::Text(tostring(G_DataState.TMIOSearchResults[i].Points) + " \\$999(#" + tostring(G_DataState.TMIOSearchResults[i].Rank) + ")");

					UI::PopID();
				}
			}

			UI::EndTable();
		}
	}
}

class FriendsTab : Tab {
	FriendsTab() {
		Label = Icons::UserCircle + " Uplay Friends";
	}

	void Render() override {
		if (!G_DataState.FriendListValid) {
			string SearchingString = "Loading...";
			vec2 SearchingSize = Draw::MeasureString(SearchingString);
			UI::SetCursorPos((UI::GetWindowSize() - SearchingSize) / 2);
			UI::Text(SearchingString);
			return;
		}

		if (UI::BeginTable("UplayFriendsList", 3, UI::TableFlags::ScrollY)) {
			UI::TableSetupScrollFreeze(0,1);
			UI::TableSetupColumn(" ", UI::TableColumnFlags::WidthFixed, (30 * UI::GetScale()));
			UI::TableSetupColumn("Friend", UI::TableColumnFlags::WidthStretch);
			UI::TableSetupColumn("Relationship", UI::TableColumnFlags::WidthFixed, UI::GetWindowSize().x * 0.3);
			UI::TableHeadersRow();

			UI::ListClipper Clipper(G_DataState.UplayFriends.Length);
			while (Clipper.Step()) {
				
				for (int i = Clipper.DisplayStart; i < Clipper.DisplayEnd; i++) {
					UI::TableNextRow();
					UI::PushID("UplayFriend" + i);

					// Button Column
					UI::TableSetColumnIndex(0);
					bool UserInParty = false;
					bool UserLocked = false;
					bool UserInvited = false;

					for (uint j = 0; j < G_QueueState.Party.CurrentSquad.Members.Length; j++) {
						if (G_DataState.UplayFriends[i].UplayPlayer.AccountID == G_QueueState.Party.CurrentSquad.Members[j].Player.AccountID) {
							UserInParty = true;
							if (G_QueueState.Party.CurrentSquad.Members[j].IsLeader || G_QueueState.Party.CurrentSquad.Members[j].Player.AccountID == GetMainUserInfo().Name) UserLocked = true;
							break;
						}
					}

					for (uint j = 0; j < G_QueueState.Party.CurrentSquad.InvitedPlayers.Length; j++) {
						if (G_DataState.UplayFriends[i].UplayPlayer.AccountID == G_QueueState.Party.CurrentSquad.InvitedPlayers[j].AccountID) {
							UserInvited = true;
							break;
						}
					}

					// Decide button color
					if (UserLocked) {
						UI::PushStyleColor(UI::Col::Button, G_UIState.Gray.Color4);
						UI::PushStyleColor(UI::Col::ButtonHovered, G_UIState.Gray.ColorHover);
						UI::PushStyleColor(UI::Col::ButtonActive, G_UIState.Gray.ColorActive);
					} else if (UserInParty) {
						UI::PushStyleColor(UI::Col::Button, G_UIState.Red.Color4);
						UI::PushStyleColor(UI::Col::ButtonHovered, G_UIState.Red.ColorHover);
						UI::PushStyleColor(UI::Col::ButtonActive, G_UIState.Red.ColorActive);
					} else if (UserInvited) {
						UI::PushStyleColor(UI::Col::Button, G_UIState.Blue.Color4);
						UI::PushStyleColor(UI::Col::ButtonHovered, G_UIState.Blue.ColorHover);
						UI::PushStyleColor(UI::Col::ButtonActive, G_UIState.Blue.ColorActive);
					} else {
						UI::PushStyleColor(UI::Col::Button, G_UIState.Green.Color4);
						UI::PushStyleColor(UI::Col::ButtonHovered, G_UIState.Green.ColorHover);
						UI::PushStyleColor(UI::Col::ButtonActive, G_UIState.Green.ColorActive);
					}

					// Decide button label / action
					if (UserInParty) {
						if (UI::Button(Icons::UserTimes)) {
							G_QueueState.Party.CurrentSquad.KickMember(G_DataState.UplayFriends[i].UplayPlayer);
						}
					} else {
						if (UI::Button(Icons::UserPlus) && !UserInvited) {
							G_QueueState.Party.CurrentSquad.InvitePlayer(G_DataState.UplayFriends[i].UplayPlayer);
						}
					}

					UI::PopStyleColor(3);

					// Name
					string PresenceString = "";
					UI::TableSetColumnIndex(1);
					if (G_DataState.UplayFriends[i].Presence == "Online") {
						PresenceString += "\\$3A5" + Icons::Circle;
					} else if (G_DataState.UplayFriends[i].Presence == "DoNotDisturb") {
						PresenceString += "\\$E44" + Icons::Circle;
					} else if (G_DataState.UplayFriends[i].Presence == "Away") {
						PresenceString += "\\$FA1" + Icons::Circle;
					} else {
						PresenceString += "\\$777" + Icons::Circle;
					}

					UI::Text(PresenceString + " \\$fff" + G_DataState.UplayFriends[i].UplayPlayer.DisplayName);

					// Relationship
					UI::TableSetColumnIndex(2);
					UI::Text(G_DataState.UplayFriends[i].Relationship + " \\$aaa(" + G_DataState.UplayFriends[i].PlatformType + ")");

					UI::PopID();
				}
			}

			UI::EndTable();
		}
	}
}

class RecentMembersTab : Tab {
	RecentMembersTab() {
		Label = Icons::ClipboardList + " Recent";
	}

	void Render() override {
		if (UI::BeginTable("RecentTeammatesList", 3, UI::TableFlags::ScrollY)) {
			UI::TableSetupScrollFreeze(0,1);
			UI::TableSetupColumn(" ", UI::TableColumnFlags::WidthFixed, (30 * UI::GetScale()));
			UI::TableSetupColumn("Player", UI::TableColumnFlags::WidthStretch);
			UI::TableSetupColumn("Last queue", UI::TableColumnFlags::WidthFixed, UI::GetWindowSize().x * 0.3);
			UI::TableHeadersRow();

			UI::ListClipper Clipper(G_DataState.RecentMembers.Length);
			while (Clipper.Step()) {
				
				for (int i = Clipper.DisplayStart; i < Clipper.DisplayEnd; i++) {
					UI::TableNextRow();
					UI::PushID("RecentMember" + i);

					// Button Column
					UI::TableSetColumnIndex(0);
					bool UserInParty = false;
					bool UserLocked = false;
					bool UserInvited = false;

					for (uint j = 0; j < G_QueueState.Party.CurrentSquad.Members.Length; j++) {
						if (G_DataState.RecentMembers[i].Player.AccountID == G_QueueState.Party.CurrentSquad.Members[j].Player.AccountID) {
							UserInParty = true;
							if (G_QueueState.Party.CurrentSquad.Members[j].IsLeader || G_QueueState.Party.CurrentSquad.Members[j].Player.AccountID == GetMainUserInfo().Name) UserLocked = true;
							break;
						}
					}

					for (uint j = 0; j < G_QueueState.Party.CurrentSquad.InvitedPlayers.Length; j++) {
						if (G_DataState.RecentMembers[i].Player.AccountID == G_QueueState.Party.CurrentSquad.InvitedPlayers[j].AccountID) {
							UserInvited = true;
							break;
						}
					}

					// Decide button color
					if (UserLocked) {
						UI::PushStyleColor(UI::Col::Button, G_UIState.Gray.Color4);
						UI::PushStyleColor(UI::Col::ButtonHovered, G_UIState.Gray.ColorHover);
						UI::PushStyleColor(UI::Col::ButtonActive, G_UIState.Gray.ColorActive);
					} else if (UserInParty) {
						UI::PushStyleColor(UI::Col::Button, G_UIState.Red.Color4);
						UI::PushStyleColor(UI::Col::ButtonHovered, G_UIState.Red.ColorHover);
						UI::PushStyleColor(UI::Col::ButtonActive, G_UIState.Red.ColorActive);
					} else if (UserInvited) {
						UI::PushStyleColor(UI::Col::Button, G_UIState.Blue.Color4);
						UI::PushStyleColor(UI::Col::ButtonHovered, G_UIState.Blue.ColorHover);
						UI::PushStyleColor(UI::Col::ButtonActive, G_UIState.Blue.ColorActive);
					} else {
						UI::PushStyleColor(UI::Col::Button, G_UIState.Green.Color4);
						UI::PushStyleColor(UI::Col::ButtonHovered, G_UIState.Green.ColorHover);
						UI::PushStyleColor(UI::Col::ButtonActive, G_UIState.Green.ColorActive);
					}

					// Decide button label / action
					if (UserInParty) {
						if (UI::Button(Icons::UserTimes)) {
							G_QueueState.Party.CurrentSquad.KickMember(G_DataState.RecentMembers[i].Player);
						}
					} else {
						if (UI::Button(Icons::UserPlus) && !UserInvited) {
							G_QueueState.Party.CurrentSquad.InvitePlayer(G_DataState.RecentMembers[i].Player);
						}
					}

					UI::PopStyleColor(3);
					UI::TableSetColumnIndex(1);

					// Player name
					UI::Text(G_DataState.RecentMembers[i].Player.DisplayName);

					// Last match Timestamp
					UI::TableSetColumnIndex(2);
					UI::Text(Time::FormatString("%x %X",G_DataState.RecentMembers[i].LastQueueStamp));
					

					UI::PopID();
				}
			}

			UI::EndTable();
		}
	}
}

// Data System

class DataState {
	bool FriendListValid;
	uint64 LastFriendListRefresh;
	array<UplayFriend@> UplayFriends;

	string TMIOSearchString;
	bool TMIOSearchResultsValid = true;
	array<LeaderboardPlayer@> TMIOSearchResults;

	array<RecentTeammate@> RecentMembers;

	string[] SquadCreationTypes = {
		"Ranked", // default for SquadCreationType
		"Royal"
	};

	

	string SquadCreationName = GetMainUserInfo().Name + "'s Squad";
	string SquadCreationType = SquadCreationTypes[0]; // string to allow arbitrary types later?

	string PartyCode = "";

	uint64 LastActivityTick;
	uint LastActivityType;

	string[] NotificationSettings = {
		"Everyone",
		"Friends only",
		"No-one"
	};
	bool NotificationsAvailable;

	string[] StatsSettings = {
		"All",
		"Past month",
		"Past week",
		"Past 24 hours",
		"Today"
	};

	void AddRecentMembers() {
		for (uint i = 0; i < G_QueueState.Party.CurrentSquad.Members.Length; i++) {
			if (G_QueueState.Party.CurrentSquad.Members[i].Player.AccountID != GetMainUserInfo().WebServicesUserId) {
				bool found = false;
				for (uint j = 0; j < RecentMembers.Length; j++) {
					if (G_QueueState.Party.CurrentSquad.Members[i].Player.AccountID == RecentMembers[j].Player.AccountID) {
						RecentMembers[j].LastQueueStamp = Time::Stamp;
						found = true;
						break;
					}
				}
				if (!found) {
					RecentMembers.InsertLast(RecentTeammate(G_QueueState.Party.CurrentSquad.Members[i].Player, Time::Stamp));
				}
			}
		}
		if (RecentMembers.Length > 1) RecentMembers.Sort(CompareRecentMembers);
		while (RecentMembers.Length > 10) RecentMembers.RemoveLast();
	}

	string GetMapNameFromURL(const string &in URL) {
		string[] Directories = URL.Split("/");
		if (Directories.Length <= 0) {
			Error(ErrorType::Error, "Map name resolver", "Cannot split URL!", URL);
			return URL;
		}
		return Directories[Directories.Length - 1].Replace(".Map.Gbx","").Replace("_"," ");
	}

	DataState() {
		startnew(GetFriendList);
	}
}

// -- Coroutine argument containers

class StringContainer {
	string s;

	StringContainer(const string &in str) {
		s = str;
	}
}

// --- Globals

QueueState@ G_QueueState;
UIState@ G_UIState;
DataState@ G_DataState;

// --- Settings

[Setting name="Notification Volume" drag min="0" max="100" description="Drag this slider to change the volume of notification sounds."]
float Setting_SoundVolume = 30.0f;

[Setting name="Use custom UI colors" description="Enable this to have the menu tabs colored after your Matchmaking rank!"]
bool Setting_ColoredUI = true;

[Setting name="Hide silent warning popups" description="These popups should only concern you if they repeatedly happen every ~5s or so. You can mute them if you want (but don't complain about stuff stopping to work without you noticing!)"]
bool Setting_HideWarn = false;

[Setting name="Auto-Accept" description="Automatically accept any incoming matches!"]
bool Setting_AutoAccept = true;

[Setting name="Show window" hidden]
bool Setting_WindowVisible = true;

[Setting name="Play sounds for invitations from" min="0" max="2" hidden]
int Setting_ShowInviteNotif = 0;

[Setting name="Show stats in the timeframe of" min="0" max="4" hidden]
int Setting_ShowPastStats = 0;

[Setting name="Receive Squadding Notifications" description="Receive Squadding Notifications through the Plugin and NOT the Nadeo menu (only one works)"]
bool Setting_ReceiveNotifications = false;

[Setting name="Highest experienced Queue Time" hidden]
uint Setting_HighestWaitingTime = 0;

// === METHODS ===

void Main() {
	// Wait for network
	NadeoServices::AddAudience("NadeoClubServices");
	while (!NadeoServices::IsAuthenticated("NadeoClubServices")) {
		yield();
	}

	// Initialize states
	@G_QueueState = QueueState();
	@G_UIState = UIState();
	@G_DataState = DataState();

	// Fill initialized states with data
	InitializeQueues();
	InitializeUI();

	// Start notification loop
	startnew(NotificationLoop);

	// Loading done?
	if (G_QueueState.Queues.Length > 0) {
		G_QueueState.Ready = true;
	} else {
		G_UIState.LoadingString = "\\$f88Loading failed! Please check the log.";
	}
}

// --- Init

void InitializeQueues() {
	if (G_UIState !is null) G_UIState.LoadingString = "Loading Queues...";

	// Completely hardcode this now thanks to API update *sigh*
	array<QueueTemplate@> KnownQueues;
	KnownQueues.InsertLast(QueueTemplate(2, "Teams 3v3"));
	KnownQueues.InsertLast(QueueTemplate(3, "Royal"));
	KnownQueues.InsertLast(QueueTemplate(4, "Super Royal"));

	for (uint i = 0; i < KnownQueues.Length; i++) {
		try {
			Queue@ NewQueue = Queue();
			NewQueue.QueueID = KnownQueues[i].QueueID;
			NewQueue.Name = KnownQueues[i].Name;
			G_QueueState.AddQueue(NewQueue);
		} catch {
			Error(ErrorType::Error, "InitializeQueues", "Unknown error while adding queue " + KnownQueues[i].Name + "!", getExceptionInfo());
		}
	}
}

void InitializeUI() {
	if (G_UIState !is null) G_UIState.LoadingString = "Loading UI Data...";

	auto app = cast<CTrackMania>(GetApp());

	// Set user color
	if (app.ManiaPlanetScriptAPI.UserMgr.Users.Length >= 1) {
		G_UIState.UserColor.SetColor(hsv2rgb(vec3(app.ManiaPlanetScriptAPI.UserMgr.Users[0].Config.User_LightTrailHue, 0.7, 0.8)));
	}

	RefreshRankColor();
}

// Init loop: someone has a match currently!
void StartQueue(ref@ q) {
	Queue@ queue = cast<Queue>(q);
	if (queue is null) {
		Error(ErrorType::Error, "StartQueue", "It looks like you have a match in progress but I can't find it, please join it manually!", "");
		return;
	}
	
	while (!G_QueueState.Ready) yield();
	@G_QueueState.CurrentQueue = queue;
	G_QueueState.IsInQueue = true;
	startnew(QueueLoop);
}

// --- Main Loops

// Notification loop: Always run, just yield if you shouldnt check
void NotificationLoop() {
	while (true) {
		try {
			RefreshNotifications();
			yield();
		} catch {
			Error(ErrorType::Error, "InvitationLoop", "Something went wrong when checking notifications!", getExceptionInfo());
		}
	}
}

// Queue Loop: Runs while queueing
void QueueLoop() {
	bool FirstHeartbeat = true;
	G_QueueState.SetStatus(QueueStatus::StartingQueue);
	G_QueueState.IsInQueue = true;
	G_QueueState.QueueTimeStart = Time::Now;
	G_QueueState.Party.RefreshPlayWith();
	try {
		G_DataState.AddRecentMembers();
		//@DEBUG SaveStats();
	} catch {
		Error(ErrorType::Error, "QueueLoop", "Couldn't save stats to keep up with recent team members.", getExceptionInfo());
	}
	while (true) {
		try {
			if (!G_QueueState.IsInQueue && !FirstHeartbeat) break;
			uint64 NextHeartbeat = Time::Now + 5000;
			SendHeartbeat();
			FirstHeartbeat = false;
			while (Time::Now < NextHeartbeat && G_QueueState.IsInQueue) yield();
		} catch {
			Error(ErrorType::Error, "QueueLoop", "Something went wrong in the queue communication!", getExceptionInfo());
		}
	}
	G_QueueState.QueueTimeStart = 0;
}

// Match Loop: Runs while in match
void MatchLoop() {
	// Get initial match info
	G_QueueState.SetStatus(QueueStatus::InMatch);
	G_QueueState.Absent = false;
	G_UIState.ShowMatchLeaveUI = true;
	// fill participant pbs
	while (true) {
		try {
			if (G_QueueState.Status != QueueStatus::InMatch) break;
			uint64 NextRefresh = Time::Now + 12000;
			CheckRoyalMatchLeft();
			G_QueueState.RefreshCurrentMatchStatus(G_QueueState.CurrentMatch.LID);
			ParseCurrentMatchStatus();
			while (Time::Now < NextRefresh && G_QueueState.Status == QueueStatus::InMatch) yield();
		} catch {
			Error(ErrorType::Log, "MatchLoop", "Something went wrong in the match status check!", getExceptionInfo());
		}
	}
}

// --- Loop tasks

void RefreshNotifications() {
	auto app = cast<CTrackMania>(GetApp());

	G_DataState.NotificationsAvailable = cast<CGameWebServicesNotificationManagerScript>(app.ManiaPlanetScriptAPI.WSNotificationMgr).Notification_IsAvailable(GetMainUserId());

	if (Setting_ReceiveNotifications) {
		while (cast<CGameWebServicesNotificationManagerScript>(app.ManiaPlanetScriptAPI.WSNotificationMgr).Notification_IsAvailable(GetMainUserId())) {
			CWebServicesTaskResult_WSNotificationScript@ nextNotificationResult = cast<CGameWebServicesNotificationManagerScript>(app.ManiaPlanetScriptAPI.WSNotificationMgr).Notification_PopNext(GetMainUserId());
			while (nextNotificationResult.IsProcessing) yield();
			if (nextNotificationResult.HasSucceeded) {
				if (nextNotificationResult.Notification.Type == "SquadInvitationReceived") {
					// @SQUADS
				} else if (nextNotificationResult.Notification.Type == "SquadInvitationCanceled") {
					// @SQUADS
				} else if (nextNotificationResult.Notification.Type == "SquadInvitationCanceledForExitingPlayer") {
					// @SQUADS
				} else if (nextNotificationResult.Notification.Type == "SquadInvitationCanceledForFullSquad") {
					// @SQUADS
				} else if (nextNotificationResult.Notification.Type == "SquadDeleted") {
					// @SQUADS
				} else if (nextNotificationResult.Notification.Type == "SquadInvitationAccepted" || nextNotificationResult.Notification.Type == "SquadInvitationDeclined" || nextNotificationResult.Notification.Type == "SquadMemberRemoved" || nextNotificationResult.Notification.Type == "SquadMemberKicked"|| nextNotificationResult.Notification.Type == "SquadInvitationAdded") {
					// @SQUADS
				// Quick hack for prestige messages also being sent over the WSNotifications to notify user and avoid error, will figure out how to avoid the "stack-popping" problem later
				} else if (nextNotificationResult.Notification.Type == "PrestigeEarned") {
					CNotification_PrestigeEarned@ prestigeEarnedResult = cast<CNotification_PrestigeEarned>(nextNotificationResult.Notification);
					if (prestigeEarnedResult is null) {
						Error(ErrorType::Log, "Notification_PopNext", "Could not resolve prestive notification result!", "");
					} else {
						UI::ShowNotification("Prestige earned!", "Category: " + prestigeEarnedResult.CategoryType + " (Level " + prestigeEarnedResult.CategoryLevel + ")\nMode: " + tostring(prestigeEarnedResult.Mode) + "\nLevel: " + prestigeEarnedResult.PrestigeLevel + "\nReward: " + prestigeEarnedResult.RewardDisplayName + "\nNext Level: " + prestigeEarnedResult.StatCurrentValue + "/" + prestigeEarnedResult.StatValueForNextLevel, 20000);
					}
				} else {
					Error(ErrorType::Log, "Notification_PopNext", "Unknown notification type! Please report this to Nsgr <3", nextNotificationResult.Notification.Type);
				}
			} else {
				Error(ErrorType::Log, "Notification_PopNext", nextNotificationResult.ErrorDescription, nextNotificationResult.ErrorCode);
			}
			cast<CGameWebServicesNotificationManagerScript>(app.ManiaPlanetScriptAPI.WSNotificationMgr).TaskResult_Release(nextNotificationResult.Id);
		}
	}
}

void NotifyInvitation(Player@ player) {
	ShowInviteNotification(player.DisplayName);
	switch (Setting_ShowInviteNotif) {
		case 0:
			startnew(PlayInviteSound);
			break;
		case 1:
			for (uint i = 0; i < G_DataState.UplayFriends.Length; i++) {
				if (G_DataState.UplayFriends[i].UplayPlayer.AccountID == player.AccountID) {
					startnew(PlayInviteSound);
					return;
				}
			}
			break;
		case 2:
			break;
		default:
			Error(ErrorType::Error, "NotifyInvitation", "Invalid Setting for Notification privacy", "" + Setting_ShowInviteNotif);
			Setting_ShowInviteNotif = 2;
	}
}

void ShowInviteNotification(const string &in InviterName) {
	UI::ShowNotification("\\$2f7" + Icons::EnvelopeOpen + " New invitation from " + InviterName, 20000);
}

const string[] FOUND_MATCH_MESSAGES = {
	"Super Secret Message", // Hello to whoever dared to open this 3000 line historically grown and very much legacy code-infested script that I want to write from scratch for months already!
	"LETS GOOOO",
	"Ready to get sniped? PepeLaugh",
	"PauseChamp",
	"I hope this didn't ruin a run lol",
	"DinkDonk",
	"Posture check",
	"Stay hydrated!",
	"Sending acceleration penalties...",
	"Building Terrain...",
	"You lost the game",
	"shitfest is real",
	// "Ping Nsgr on the Openplanet Discord to continue working on AFK Tool v3 already",
	"xdd",
	"Joel"
};

string GetMatchReadyText() {
	if (Math::Rand(0., 1.) > 0.95) {
		return FOUND_MATCH_MESSAGES[Math::Rand(0, FOUND_MATCH_MESSAGES.Length - 1)];
	}
	return "You will join automatically from the main menu once the server responds.";
}

void SendHeartbeat() {
	G_QueueState.LastHeartbeatSent = Time::Now;
	string HeartbeatResponse = SendNadeoRequest(Net::HttpMethod::Post, "https://meet.trackmania.nadeo.club/api/matchmaking/" + G_QueueState.CurrentQueue.QueueID + "/heartbeat", G_QueueState.Party.PlayWith);
	Json::Value HeartbeatResult = ResponseToJSON(HeartbeatResponse, Json::Type::Object);
	if (!G_QueueState.IsInQueue) {
		Error(ErrorType::Log, "Queue communication", "Match canceled between heartbeat send and receive...", HeartbeatResponse);
		return;
	}
	if (HeartbeatResult.GetType() != Json::Type::Null && HeartbeatResult.HasKey("status")) {
		G_QueueState.LastHeartbeat = Time::Now;
		QueueStatus ReceivedStatus = GetStatusFromString(HeartbeatResult["status"]);
		string BanInfoString = "BONK!";
		switch (ReceivedStatus) {
			case QueueStatus::Pending :
			case QueueStatus::Queued :
			case QueueStatus::MatchAccepted :
				break;
			case QueueStatus::Canceled :
				if (Time::Now > G_QueueState.MatchFoundTime + C_MatchFoundTimeout && G_QueueState.Status == QueueStatus::MatchFound) {
					if (G_QueueState.IncreaseBonkCounter()) return;
				}
				break;
			case QueueStatus::SkillGapCanceled :
				G_QueueState.IsInQueue = false;
				break;
			case QueueStatus::Banned :
				G_QueueState.IsInQueue = false;
				if (HeartbeatResult.HasKey("banEndDate") && HeartbeatResult["banEndDate"].GetType() == Json::Type::Number) {
					BanInfoString = "You ban ends in " + Time::Format((uint(HeartbeatResult["banEndDate"]) - Time::Stamp) * 1000) + ".";
				}
				Error(ErrorType::UserWarning, "You are banned from the queue!", BanInfoString, HeartbeatResponse);
				break;
			case QueueStatus::MatchFound :
				// This does not really happen anymore, but in case it ever returns we keep handling this.
				if (G_QueueState.Status != QueueStatus::MatchFound) {
					G_QueueState.MatchFoundTime = Time::Now;
					G_QueueState.Status = QueueStatus::MatchFound;
					startnew(FoundMatch);
				}
				break;
			case QueueStatus::MatchReady :
				if (G_QueueState.Status != QueueStatus::MatchReady && G_QueueState.Status != QueueStatus::MatchAccepted && G_QueueState.Status != QueueStatus::MatchAccepted) {
					cast<CTrackMania>(GetApp()).ManiaPlanetScriptAPI.FlashWindow();
					UI::ShowNotification("\\$2f7" + Icons::SignInAlt + " Match found!", GetMatchReadyText(), 10000);
					startnew(PlayMatchFoundSound);
				}
				if (HeartbeatResult.HasKey("matchLiveId") && HeartbeatResult["matchLiveId"].GetType() != Json::Type::Null) {
					G_QueueState.IsInQueue = false;
					G_QueueState.MatchFoundTime = 0;
					@G_QueueState.CurrentMatch = Match();
					
					// Get initial match loop
					while (G_QueueState.CurrentMatch.LID == "") {
						G_QueueState.RefreshCurrentMatchStatus(HeartbeatResult["matchLiveId"], true);
						if (G_QueueState.CurrentMatch.LID == "") sleep(5000);
					}
					startnew(JoinMatch);
					G_QueueState.CurrentMatch.SetupPlayerData();
					return;
				}
				break;
			default:
				Error(ErrorType::Error, "Queue communication", "Unknown heartbeat status!", HeartbeatResponse);
		}
		G_QueueState.SetStatus(ReceivedStatus);
	} else {
		Error(ErrorType::Warn, "Queue communication", "Invalid response from server!", HeartbeatResponse);
	}
}

void FoundMatch() {
	cast<CTrackMania>(GetApp()).ManiaPlanetScriptAPI.FlashWindow();
	G_QueueState.InputRequired = false;
	if (Setting_AutoAccept) {
		if (Time::Now < GetMostRecentInputTick() + 10000) {
			UI::ShowNotification("\\$2f7" + Icons::SignInAlt + " Match found and accepted!", "PauseChamp", TimeUntilAcceptTimeout());
			startnew(SendAccept);
			startnew(PlayMatchFoundSound);
			return;
		} else {
			UI::ShowNotification("\\$2f7" + Icons::SignInAlt + " Match found, but are you AFK?", "PLEASE PROVIDE INPUT TO CONFIRM YOU ARE THERE", TimeUntilAcceptTimeout());
			startnew(PlayMatchFoundInputSound);
		}
	} else {
		UI::ShowNotification("\\$2f7" + Icons::SignInAlt + " Match found!", "Please Accept in the overlay!", TimeUntilAcceptTimeout());
		startnew(PlayMatchFoundInputSound);
		return;
	}
	G_QueueState.InputRequired = true;
	while (G_QueueState.Status == QueueStatus::MatchFound && Time::Now < G_QueueState.MatchFoundTime + C_MatchFoundTimeout + C_MatchFoundTimeout) {
		if (Time::Now < GetMostRecentInputTick() + 5000) {
			startnew(SendAccept);
			string NotificationString = "Received ";
			switch (G_DataState.LastActivityType) {
				case 0:
					NotificationString += "keyboard";
					break;
				case 1:
					NotificationString += "mouse";
					break;
				case 2:
					NotificationString += "pad";
					break;
				default:
					NotificationString += "unknown";
			}
			NotificationString += " input!";
			UI::ShowNotification("\\$2f7" + Icons::SignInAlt + " Match accepted!", NotificationString, TimeUntilAcceptTimeout());
			startnew(PlayMatchFoundInputReceivedSound);
			G_QueueState.InputRequired = false;
			return;
		}
		yield();
	}
}

uint64 GetMostRecentInputTick() {
	uint64 MostRecentPadDelay = Time::Now;
	auto InputPort = cast<CTrackMania>(GetApp()).InputPort;
	for (uint i = 0; i < InputPort.Script_Pads.Length; i++) {
		MostRecentPadDelay = Math::Min(MostRecentPadDelay, InputPort.Script_Pads[i].IdleDuration);
	}
	return Math::Max(G_DataState.LastActivityTick, Time::Now - MostRecentPadDelay);
}

int TimeUntilAcceptTimeout() {
	return G_QueueState.MatchFoundTime + C_MatchFoundTimeout - Time::Now;
}

void SendAccept() {
	SendNadeoRequest(Net::HttpMethod::Post, "https://meet.trackmania.nadeo.club/api/matchmaking/" + G_QueueState.CurrentQueue.QueueID + "/accept");
}

void SendCancel() {
	SendNadeoRequest(Net::HttpMethod::Post, "https://meet.trackmania.nadeo.club/api/matchmaking/" + G_QueueState.CurrentQueue.QueueID + "/cancel");
}

void PlayerStatusCheck(ref@ q) {
	Queue@ queue = cast<Queue>(q);
	if (queue is null) {
		Error(ErrorType::Error, "PlayerStatusCheck", "No queue passed as argument!", "");
		return;
	}
	G_QueueState.RefreshPlayerStatus(queue);
}

void JoinMatch() {
	G_QueueState.SetStatus(QueueStatus::WaitForServer);
	auto app = cast<CTrackMania>(GetApp());

	uint64 CutOffJoinAnyway = Time::Now + 20000;
	uint64 NextJoinAttempt = 0;
	uint64 NextCheckAttempt = 0;
	uint64 NextBackToMenu = 0;
	uint64 NextRefreshAttempt = 0;

	while (!IsInMatch()) {
		if (Time::Now > CutOffJoinAnyway && Time::Now > NextJoinAttempt) {
			NextJoinAttempt = Time::Now + 3000;
			JoinServer();
		}

		if (G_QueueState.CurrentMatch.JoinLink.Length > 0) {
			if (Time::Now <= CutOffJoinAnyway) {
				if (app.ManiaTitleControlScriptAPI.IsReady) {
					if (Time::Now > NextCheckAttempt) {
						NextCheckAttempt = Time::Now + 2000;
						if (IsServerReady(G_QueueState.CurrentMatch.JoinLink.Replace("#qjoin=","").Replace("#join=","").Replace("@Trackmania",""), CutOffJoinAnyway)) {
							NextJoinAttempt = Time::Now + 3000;
							JoinServer();
						}
					}
				} else {
					if (Time::Now > NextBackToMenu && G_QueueState.Status == QueueStatus::WaitForServer) {
						NextBackToMenu = Time::Now + 5000;
						app.BackToMainMenu();
					}
				}
			}
		} else if (Time::Now > NextRefreshAttempt) {
			NextRefreshAttempt = Time::Now + 5000;
			G_QueueState.RefreshCurrentMatchStatus(G_QueueState.CurrentMatch.LID);
		}

		yield();
	}
	startnew(MatchLoop);
}

bool IsServerReady(const string &in ServerLogin, uint64 CutOffTime) {
	auto app = cast<CTrackMania>(GetApp());
	while (Time::Now < CutOffTime && !app.ManiaTitleControlScriptAPI.IsReady) yield();
	if (app.ManiaTitleControlScriptAPI.IsReady) {
		app.ManiaTitleControlScriptAPI.GetServerInfo(ServerLogin);
		yield(); // Yield to make sure the title control can switch to being not ready
		while (!app.ManiaTitleControlScriptAPI.IsReady && Time::Now < CutOffTime) yield();
		if (app.ManiaTitleControlScriptAPI.IsReady) {
			return (app.ManiaTitleControlScriptAPI.LatestResult == CGameManiaTitleControlScriptAPI::EResult::Success);
		}
	}
	
	return true;
}

void JoinServer() {
	if (G_QueueState.Status == QueueStatus::WaitForServer) G_QueueState.SetStatus(QueueStatus::Joining);
	auto app = cast<CTrackMania>(GetApp());
	if (app.ManiaPlanetScriptAPI.ActiveContext_InGameMenuDisplayed) {
		app.CurrentPlayground.Interface.ManialinkScriptHandler.CloseInGameMenu(CGameScriptHandlerPlaygroundInterface::EInGameMenuResult::Resume);
		sleep(250);
		if (app.ManiaPlanetScriptAPI.ActiveContext_InGameMenuDisplayed) {
			Error(ErrorType::UserWarning, "Please close the Pause menu!", "Softlock prevention: Please exit first!", "");
			return;
		}
	}
	if (!IsInMatch()) {
		app.ManiaPlanetScriptAPI.OpenLink(G_QueueState.CurrentMatch.JoinLink.Replace("#join=","#qjoin="), CGameManiaPlanetScriptAPI::ELinkType::ManialinkBrowser);
	}
}

bool IsInMatch() {
	auto ServerInfo = cast<CGameCtnNetServerInfo>(cast<CTrackManiaNetwork>(GetApp().Network).ServerInfo);
	return ServerInfo.ServerLogin != "" && ServerInfo.ServerLogin == G_QueueState.CurrentMatch.JoinLink.Replace("#qjoin=","").Replace("#join=","").Replace("@Trackmania","");
}

void CheckRoyalMatchLeft() {
	// @ROYAL
	if (G_QueueState.CurrentQueue.Stats.CurrentDivision.RuleType == DivisionRuleType::VictoryCount) {
		if (!IsInMatch()) {
			G_QueueState.Absent = true;
		} else {
			G_QueueState.Absent = false;
			G_UIState.ShowMatchLeaveUI = true;
		}
	}
}

void LeaveMatch() {
	G_QueueState.CurrentQueue.Stats.CurrentRanking = RefreshQueueRanking(G_QueueState.CurrentQueue.QueueID);
	G_QueueState.CurrentQueue.Stats.RefreshDivisionRanking();
	RefreshRankColor();
	G_QueueState.SetStatus(QueueStatus::MatchCompleted);
	@G_QueueState.CurrentMatch = null;
	G_QueueState.IsInQueue = false;
}

void ParseCurrentMatchStatus() {
	if (G_QueueState.CurrentMatch.Status == "COMPLETED") {
		G_QueueState.CurrentQueue.Stats.CurrentRanking = RefreshQueueRanking(G_QueueState.CurrentQueue.QueueID);
		G_QueueState.CurrentQueue.Stats.RefreshDivisionRanking();
		RefreshRankColor();
		G_QueueState.SetStatus(QueueStatus::MatchCompleted);
		G_QueueState.CurrentQueue.Stats.SaveMatchForStats(G_QueueState.CurrentMatch);
		@G_QueueState.CurrentMatch = null;

		if (G_QueueState.CurrentQueue.QueueID == 4) {
			Error(ErrorType::Info, "", "End of Super Royal match!", "To continue playing Super Royal, you will need to use the Nadeo menu's methods.");
		}

		if (G_QueueState.AutoRequeue) {
			G_QueueState.IsInQueue = true;
			startnew(QueueLoop);
			Error(ErrorType::Info, "", "Match ended!", "You have been automatically requeued!");
			G_QueueState.AutoRequeue = false;
		} else {
			G_QueueState.IsInQueue = false;
		}
	}
}

void ExitQueue() {
	G_QueueState.IsInQueue = false;
	startnew(SendCancel);
	G_QueueState.SetStatus(QueueStatus::CanceledUser);
}

void CopyMatchID() {
	if (G_QueueState.CurrentMatch !is null) IO::SetClipboard(G_QueueState.CurrentMatch.LID);
}

// --- UI

void RenderMenu() {
	if (UI::MenuItem(CreateColorCode(hsv2rgb(vec3(Math::Fract(Time::Now / 10000.0), 0.4, 1.0))) + Icons::Gamepad + "  \\$zAFK Queue Tool", "", Setting_WindowVisible)) {
		Setting_WindowVisible = !Setting_WindowVisible;
		if (G_UIState !is null) G_UIState.ResizeMainWindow = true;
	}
}

void RenderInterface() {
	if (!Setting_WindowVisible) return;
	if (G_UIState !is null && G_UIState.ResizeMainWindow) {
		UI::SetNextWindowSize(465,364, UI::Cond::Always);
		G_UIState.ResizeMainWindow = false;
	}
	if (UI::Begin(Icons::Gamepad + "  AFK Queue Tool \\$555 by Nsgr###AFKQueue2", Setting_WindowVisible, UI::WindowFlags(1 << 21) | UI::WindowFlags::NoCollapse)) {
		if (G_UIState is null || !G_QueueState.Ready) {
			string LoadingString = (G_UIState is null ? "Loading States..." : G_UIState.LoadingString);
			vec2 LoadingSize = Draw::MeasureString(LoadingString);
			UI::SetCursorPos((UI::GetWindowSize() - LoadingSize) / 2);
			BoldText(LoadingString);
			UI::End();
			return;
		}

		// Rank header
		RenderRankBar();

		// Main Menu
		bool ApplyStyle = Setting_ColoredUI;
		if (ApplyStyle) UI::PushStyleColor(UI::Col::TabActive, G_UIState.TabBarBottom);
		UI::BeginTabBar("AFKMainTabBar", UI::TabBarFlags::FittingPolicyResizeDown);
		vec2 tabSize = vec2(UI::GetWindowSize().x - (16 * UI::GetScale()), UI::GetWindowSize().y - UI::GetCursorPos().y - (36 * UI::GetScale()));

		if (ApplyStyle) {
			UI::PushStyleColor(UI::Col::Tab, G_UIState.RankColor.Color4);
			UI::PushStyleColor(UI::Col::TabHovered, G_UIState.RankColor.ColorHover);
			UI::PushStyleColor(UI::Col::TabActive, G_UIState.RankColor.ColorActive);
		}
		
		// Render Tabs
		for (uint i = 0; i < G_UIState.Tabs.Length; i++) {
			if (UI::BeginTabItem(G_UIState.Tabs[i].Label)) {
				UI::BeginChild("AFKTab" + i, tabSize);

				G_UIState.Tabs[i].Render();

				UI::EndChild();
				UI::EndTabItem();
			}
		}
		if (ApplyStyle) UI::PopStyleColor(3);
		UI::EndTabBar();
		if (ApplyStyle) UI::PopStyleColor(1);

		// "Status bar"
		RenderStatusBar();
		UI::Dummy(vec2(0,0));
		UI::End();
	}
	
	if (!G_UIState.InviteMenuVisible) return;
	UI::SetNextWindowSize(400,250, UI::Cond::Appearing);
	if (UI::Begin(Icons::Envelope + " Add Players \\$555 to Party###AFKQueue2Invite", G_UIState.InviteMenuVisible, UI::WindowFlags(1 << 21) | UI::WindowFlags::NoCollapse)) {
		UI::BeginTabBar("AFKInviteCategory");

		for (uint i = 0; i < G_UIState.InviteTabs.Length; i++) {
			if (UI::BeginTabItem(G_UIState.InviteTabs[i].Label)) {
				G_UIState.InviteTabs[i].Render();
				UI::EndTabItem();
			}
		}

		UI::EndTabBar();
		UI::End();
	}
}

// --- UI Overlays

void RenderRankBar() {
	vec2 pos_orig = UI::GetCursorPos();
	pos_orig.y -= (2 * UI::GetScale());

	// Draw rank image
	UI::Texture@ RankImage;
	UI::Texture@ HoverImage;
	vec2 RankSize = vec2(0,0);
	if (G_QueueState.CurrentQueue.Stats.CurrentDivision !is null) {
		@RankImage = GetDivisionImage(G_QueueState.CurrentQueue.Stats.CurrentDivision.Position, G_QueueState.CurrentQueue.QueueID, false);
		@HoverImage = GetDivisionImage(G_QueueState.CurrentQueue.Stats.CurrentDivision.Position, G_QueueState.CurrentQueue.QueueID, true);
		RankSize = RankImage.GetSize();
		RankSize /= RankSize.y / (52 * UI::GetScale());
		UI::SetCursorPos(pos_orig);
		UI::Image(RankImage, RankSize);
		if (UI::IsItemHovered()) {
			float HoverSize = UI::GetWindowSize().y;
			UI::BeginTooltip();
			UI::Text(GetDivisionText(G_QueueState.CurrentQueue.Stats.CurrentDivision.Position, G_QueueState.CurrentQueue.QueueID));
			UI::Image(HoverImage, vec2(HoverSize, HoverSize));
			UI::EndTooltip();
		}
	}

	// Draw ranking info texts
	string RankString = "\\$eeeRank: " + G_QueueState.CurrentQueue.Stats.CurrentRanking.Rank + " / " + G_QueueState.CurrentQueue.Stats.CurrentRanking.Cardinal;
	string PointsString = "";
	// @ROYAL
	if (G_QueueState.CurrentQueue.Stats.CurrentDivision.RuleType == DivisionRuleType::VictoryCount) {
		PointsString = "\\$eeeWins: " + G_QueueState.CurrentQueue.Stats.CurrentRanking.Wins + " \\$999(" + G_QueueState.CurrentQueue.Stats.CurrentRanking.Points + " pts.)";
	} else {
		PointsString = "\\$eeePoints: " + G_QueueState.CurrentQueue.Stats.CurrentRanking.Points;
	}

	UI::SetCursorPos(vec2(pos_orig.x + RankSize.x + (8 * UI::GetScale()), pos_orig.y + (7 * UI::GetScale())));
	UI::Text(PointsString);

	UI::SetCursorPos(vec2(pos_orig.x + RankSize.x + (8 * UI::GetScale()), pos_orig.y + (29 * UI::GetScale())));
	UI::Text(RankString);
	if (G_QueueState.Queues.Length > 0) {
		
		UI::SetCursorPos(vec2(3 * UI::GetWindowSize().x / 5, pos_orig.y + (6 * UI::GetScale())));
		UI::SetNextItemWidth(2 * UI::GetWindowSize().x / (5 * UI::GetScale()) - 24);
		UI::PushStyleColor(UI::Col::PopupBg, vec4(0.1,0.1,0.1,1.0));
		UI::PushStyleColor(UI::Col::Border, vec4(0.4,0.4,0.4,1.0));
		if (UI::BeginCombo("##CurrentQueueCombo", G_QueueState.CurrentQueue.Name, UI::ComboFlags::NoArrowButton)) {
			for (uint i = 0; i < G_QueueState.Queues.Length; i++) {
				if (UI::Selectable(G_QueueState.Queues[i].GetTitle(), G_QueueState.Queues[i].QueueID == G_QueueState.CurrentQueue.QueueID) && !G_QueueState.IsInQueue) {
					@G_QueueState.CurrentQueue = G_QueueState.Queues[i];
					RefreshRankColor();
				}
			}
			UI::EndCombo();
		}
		UI::PopStyleColor(2);

		if (G_QueueState.Queues.Length > 1) {
			UI::SetCursorPos(vec2(UI::GetWindowSize().x - (48 * UI::GetScale()), pos_orig.y + (11 * UI::GetScale())));
			UI::Text(Icons::ChevronDown);
		}
	}

	// Set cursor for the rest
	pos_orig.y += (56 * UI::GetScale());
	UI::SetCursorPos(pos_orig);
}

void RenderStatusBar() {
	vec2 pos_orig = UI::GetCursorPos();
	float lowestPossibleTextHeight = UI::GetWindowSize().y - (25 * UI::GetScale());

	string VersionString = "\\$777v" + Meta::ExecutingPlugin().Version;
	vec2 textSize = Draw::MeasureString(VersionString);

	UI::SetCursorPos(vec2(pos_orig.x, lowestPossibleTextHeight - (5 * UI::GetScale())));
	UI::Separator();
	UI::Text(G_QueueState.GetStatusBarString()); UI::SameLine(); UI::PushFont(G_UIState.Bold); UI::Text(G_QueueState.StatusString); UI::PopFont();

	UI::SetCursorPos(vec2(UI::GetWindowSize().x - textSize.x - (8 * UI::GetScale()), lowestPossibleTextHeight));
	UI::Text(VersionString);

	UI::SetCursorPos(pos_orig);
}

// --- UI helpers

void BoldText(const string &in Text) {
	if (G_UIState !is null && G_UIState.Bold !is null) {
		UI::PushFont(G_UIState.Bold);
		UI::Text(Text);
		UI::PopFont();
	} else {
		UI::Text(Text);
	}
}

// --- UI Data

vec3 hsv2rgb(const vec3 &in c) {
	vec4 K = vec4(1.0, 2.0 / 3.0, 1.0 / 3.0, 3.0);
	vec3 p = Math::Abs(Math::Fract(vec3(c.x,c.x,c.x) + vec3(K.x,K.y,K.z)) * 6.0 - vec3(K.w,K.w,K.w));
	return Math::Lerp(vec3(K.x,K.x,K.x), Math::Clamp(p - vec3(K.x,K.x,K.x), 0.0, 1.0), c.y) * c.z;
}

string CreateColorCode(const vec3 &in Input) {
	string ReturnedValue = "\\$";
	ReturnedValue += Text::Format("%x", uint8(Input.x * 15.0));
	ReturnedValue += Text::Format("%x", uint8(Input.y * 15.0));
	ReturnedValue += Text::Format("%x", uint8(Input.z * 15.0));
	return ReturnedValue;
}

// --- Sound

enum Sounds {
	MatchFound,
	MatchFoundAwaitInput,
	MatchFoundInputReceived,
	Warn,
	Error,
	Info,
	RankUp,
	RankDown,
	InvitationReceived
}

void PlaySound(Sounds soundType) {
	float Pitch = 1.0;
	string FileName = "Race3";
	switch (soundType) {
		case Sounds::MatchFound :
			Pitch = 1.0;
			FileName = "MatchFound";
			break;
		case Sounds::MatchFoundAwaitInput :
			Pitch = 0.5;
			FileName = "MatchFound";
			break;
		case Sounds::MatchFoundInputReceived :
			Pitch = 2.0;
			FileName = "PlayButtonSelect";
			break;
		case Sounds::Warn :
			Pitch = 1.8;
			FileName = "RaceWooshStart";
			break;
		case Sounds::Error :
			Pitch = 0.5;
			FileName = "Race3";
			break;
		case Sounds::Info :
			Pitch = 1.5;
			FileName = "Race3";
			break;
		case Sounds::RankUp :
			Pitch = 1.0;
			FileName = "RankUp";
			break;
		case Sounds::RankDown :
			Pitch = 1.0;
			FileName = "RankDown";
			break;
		case Sounds::InvitationReceived :
			Pitch = 1.0;
			FileName = "InvitationNotification";
			break;
	}

	Audio::Sample@ sample = cast<Audio::Sample>(G_UIState.SoundSamples[FileName]);
	if (sample is null) {
		Error(ErrorType::Warn, "PlaySound", "Audio sample could not be casted!", FileName);
		return;
	}
	Audio::Voice@ voice = Audio::Play(sample, Setting_SoundVolume / 100.0f);
}

// --- Network Base

string SendNadeoRequest(const Net::HttpMethod Method, const string &in URL, const string &in Body = "") {
	Net::HttpRequest@ req = NadeoServices::Request("NadeoClubServices");
	req.Method = Method;
	req.Url = URL;
	req.Headers["Accept"] = "application/json";
	req.Headers["Content-Type"] = "application/json";
	req.Body = Body;
	req.Start();
	while (!req.Finished()) {
		yield();
	}
	return req.String();
}

string SendJSONRequest(const Net::HttpMethod Method, const string &in URL, const string &in Body = "") {
	dictionary@ Headers = dictionary();
	Headers["Accept"] = "application/json";
	Headers["Content-Type"] = "application/json";
	return SendHTTPRequest(Method, URL, Body, Headers);
}

string SendHTTPRequest(const Net::HttpMethod Method, const string &in URL, const string &in Body, dictionary@ Headers) {
	Net::HttpRequest req;
	req.Method = Method;
	req.Url = URL;
	@req.Headers = Headers;
	req.Body = Body;
	req.Start();
	while (!req.Finished()) {
		yield();
	}
	return req.String();
}

Json::Value ResponseToJSON(const string &in HTTPResponse, Json::Type ExpectedType) {
	Json::Value ReturnedObject;
	try {
		ReturnedObject = Json::Parse(HTTPResponse);
	} catch {
		Error(ErrorType::Error, "ResponseToJSON", "JSON Parsing of string failed!", getExceptionInfo() + "\n" + HTTPResponse);
	}
	if (ReturnedObject.GetType() != ExpectedType) {
		Error(ErrorType::Warn, "ResponseToJSON", "Unexpected JSON Type returned", getExceptionInfo() + "\n" + HTTPResponse);
		return ReturnedObject;
	}
	return ReturnedObject;
}

// --- Network Methods

float[] GetPointsHistory(const string &in AccountID) {
	float[] ReturnedValue = {C_Invalid};
	while (G_QueueState.CurrentQueue is null) yield();
	string MatchHistoryResponse = SendJSONRequest(Net::HttpMethod::Get, "https://trackmania.io/api/player/" + AccountID + "/matches/" + G_QueueState.CurrentQueue.QueueID + "/0");
	Json::Value MatchHistory = ResponseToJSON(MatchHistoryResponse, Json::Type::Object);
	if (MatchHistory.GetType() != Json::Type::Null && MatchHistory.HasKey("matches") && MatchHistory["matches"].Length > 1) {
		try {
			ReturnedValue = {};
			for (uint i = MatchHistory["matches"].Length; i > 0; i--) {
				ReturnedValue.InsertLast(MatchHistory["matches"][i-1]["afterscore"]);
			}
		} catch {
			ReturnedValue = {C_Invalid}; // need to set back in case error happened while array didnt have 2 elements yet
			Error(ErrorType::Error, "GetPointsHistory", "Cannot read past match deltas", getExceptionInfo() + "\n" + MatchHistoryResponse);
		}
	}
	return ReturnedValue;
}

// --- Data

int64 MidnightTimestamp() {
	string LocalTimeString = GetApp().OSLocalTime; //HH:MM:SS
	string[] TimeParts = LocalTimeString.Split(":"); // [HH,MM,SS]
	int64 ReturnedValue = Time::Stamp - (Text::ParseInt(TimeParts[0]) * 3600) - (Text::ParseInt(TimeParts[1]) * 60) - (Text::ParseInt(TimeParts[2]));
	return ReturnedValue;
}

QueueStatus GetStatusFromString(const string &in status) {
	if (status == "") {
		return QueueStatus::NotInQueue;
	} else if (status == "pending") {
		return QueueStatus::Pending;
	} else if (status == "skill_gap_canceled") {
		return QueueStatus::SkillGapCanceled;
	} else if (status == "different_code_canceled") {
		return QueueStatus::DifferentCodeCanceled;
	} else if (status == "crossplay_canceled") {
		return QueueStatus::CrossplayCanceled;
	} else if (status == "old_game_version") {
		return QueueStatus::OldGameVersionCanceled;
	} else if (status == "queued") {
		return QueueStatus::Queued;
	} else if (status == "match_found") {
		return QueueStatus::MatchFound;
	} else if (status == "match_accepted") {
		return QueueStatus::MatchAccepted;
	} else if (status == "match_ready") {
		return QueueStatus::MatchReady;
	} else if (status == "canceled") {
		return QueueStatus::Canceled;
	} else if (status == "banned") {
		return QueueStatus::Banned;
	} else {
		return QueueStatus::Unknown;
	}
}

void TMIOSearch() {
	G_DataState.TMIOSearchResultsValid = false;
	G_DataState.TMIOSearchResults.RemoveRange(0,G_DataState.TMIOSearchResults.Length);
	string SearchResponse = SendJSONRequest(Net::HttpMethod::Get, "https://trackmania.io/api/players/find?search=" + G_DataState.TMIOSearchString);
	Json::Value SearchResult = ResponseToJSON(SearchResponse, Json::Type::Array);
	array<LeaderboardPlayer@> Results;
	if (SearchResult.GetType() != Json::Type::Null) {
		try {
			for (uint i = 0; i < SearchResult.Length; i++) {
				try {
					if (SearchResult[i]["matchmaking"].Length > 0) {
						int rank = 0;
						int score = 0;
						for (uint j = 0; j < SearchResult[i]["matchmaking"].Length; j++) {
							if (SearchResult[i]["matchmaking"][j]["typeid"] == G_QueueState.CurrentQueue.QueueID) {
								rank = SearchResult[i]["matchmaking"][j]["rank"];
								score = SearchResult[i]["matchmaking"][j]["score"];
								break;
							}
						}
						Results.InsertLast(LeaderboardPlayer(rank, score, Player(SearchResult[i]["player"]["id"], SearchResult[i]["player"]["name"])));
					} else {
						Results.InsertLast(LeaderboardPlayer(C_Invalid, C_Invalid, Player(SearchResult[i]["player"]["id"], SearchResult[i]["player"]["name"])));
					}
				} catch {
					Error(ErrorType::Warn, "TMIO Search", "Something went wrong while parsing a search result element!", getExceptionInfo() + "\n" + SearchResponse);
				}
			}
		} catch {
			Error(ErrorType::Error, "TMIO Search", "Something went wrong while parsing the search results!", getExceptionInfo() + "\n" + SearchResponse);
		}
	} else {
		Error(ErrorType::Warn, "TMIO Search", "Received empty / invalid search result", SearchResponse);
	}

	G_DataState.TMIOSearchResults = Results;
	G_DataState.TMIOSearchResultsValid = true;
}

MwId GetMainUserId() {
	auto app = cast<CTrackMania>(GetApp());
	if (app.ManiaPlanetScriptAPI.UserMgr.MainUser !is null) {
		return app.ManiaPlanetScriptAPI.UserMgr.MainUser.Id;
	}
	if (app.ManiaPlanetScriptAPI.UserMgr.Users.Length >= 1) {
		// unexpected case but Nadeo watches out for this for some weird reason, so let's handle it
		if (app.ManiaPlanetScriptAPI.UserMgr.Users.Length > 1) {
			string ErrorString = "User buffer length: " + app.ManiaPlanetScriptAPI.UserMgr.Users.Length + "\n";
			for (uint i = 0; i < app.ManiaPlanetScriptAPI.UserMgr.Users.Length; i++) {
				ErrorString += "User #" + i + "\nDisplayName: " + app.ManiaPlanetScriptAPI.UserMgr.Users[i].DisplayName + "\nId: " + app.ManiaPlanetScriptAPI.UserMgr.Users[i].Id.GetName() + "\n";
			}
			Error(ErrorType::Warn, "UserMgr Users", "The game user list was not just one user!", ErrorString);
		}
		return app.ManiaPlanetScriptAPI.UserMgr.Users[0].Id;
	} else {
		Error(ErrorType::Error, "UserMgr Users", "The game user list is empty!", "MainUser null? :" + (app.ManiaPlanetScriptAPI.UserMgr.MainUser is null));
		return MwId();
	}
}

void GetFriendList() {
	if (Time::Now < G_DataState.LastFriendListRefresh + 10000) return;
	G_DataState.FriendListValid = false;
	G_DataState.LastFriendListRefresh = Time::Now;

	array<UplayFriend@> FriendList;
	auto app = cast<CTrackMania>(GetApp());
	
	CWebServicesTaskResult_FriendListScript@ friendListResult = app.ManiaPlanetScriptAPI.UserMgr.Friend_GetList(GetMainUserId());
	while (friendListResult.IsProcessing) yield();
	if (friendListResult.HasSucceeded) {
		MwFastBuffer<CFriend@> UplayFriendList = friendListResult.FriendList;
		for (uint i = 0; i < UplayFriendList.Length; i++) {
			FriendList.InsertLast(UplayFriend(UplayFriendList[i]));
		}
	} else {
		Error(ErrorType::Error, "Friend_GetList", friendListResult.ErrorDescription, friendListResult.ErrorCode);
	}
	app.ManiaPlanetScriptAPI.UserMgr.TaskResult_Release(friendListResult.Id);

	if (FriendList.Length > 0) FriendList.Sort(CompareUplayFriends);

	G_DataState.UplayFriends = FriendList;
	G_DataState.FriendListValid = true;
}

const dictionary DIVISION_COLOR = {
	{"0", vec3(0,0,0)},
	{"1", vec3(0.761, 0.506, 0.333)},
	{"2", vec3(0.761, 0.506, 0.333)},
	{"3", vec3(0.761, 0.506, 0.333)},
	{"4", vec3(0.682, 0.690, 0.690)},
	{"5", vec3(0.682, 0.690, 0.690)},
	{"6", vec3(0.682, 0.690, 0.690)},
	{"7", vec3(0.8745, 0.663, 0.086)},
	{"8", vec3(0.8745, 0.663, 0.086)},
	{"9", vec3(0.8745, 0.663, 0.086)},
	{"10", vec3(0.502, 0.871, 0.035)},
	{"11", vec3(0.502, 0.871, 0.035)},
	{"12", vec3(0.502, 0.871, 0.035)},
	{"13", vec3(0.502, 0.871, 0.0352)}
};

const dictionary ROYAL_DIVISION_COLOR = {
	{"0", vec3(0,0,0)},
	{"1", vec3(0.871, 0.502, 0.0352)},
	{"2", vec3(0.761, 0.506, 0.333)},
	{"3", vec3(0.682, 0.690, 0.690)},
	{"4", vec3(0.8745, 0.663, 0.086)},
	{"5", vec3(0.502, 0.871, 0.0352)}
};

const dictionary DIVISION_SHORT = {
	{"0", "???"},
	{"1","[I]"},
	{"2","[II]"},
	{"3","[III]"},
	{"4","[I]"},
	{"5","[II]"},
	{"6","[III]"},
	{"7","[I]"},
	{"8","[II]"},
	{"9","[III]"},
	{"10","[I]"},
	{"11","[II]"},
	{"12","[III]"},
	{"13","[TM]"}
};

const dictionary DIVISION_NAME = {
	{"0", "???"},
	{"1","Bronze I"},
	{"2","Bronze II"},
	{"3","Bronze III"},
	{"4","Silver I"},
	{"5","Silver II"},
	{"6","Silver III"},
	{"7","Gold I"},
	{"8","Gold II"},
	{"9","Gold III"},
	{"10","Master I"},
	{"11","Master II"},
	{"12","Master III"},
	{"13","Trackmaster"}
};

const dictionary ROYAL_DIVISION_SHORT = {
	{"0", "???"},
	{"1","[U]"},
	{"2","[B]"},
	{"3","[S]"},
	{"4","[G]"},
	{"5","[M]"}
};

const dictionary ROYAL_DIVISION_NAME = {
	{"0", "???"},
	{"1","Unranked"},
	{"2","Bronze"},
	{"3","Silver"},
	{"4","Gold"},
	{"5","Master"}
};

vec3 GetDivisionColor(uint position, uint QueueID) {
	vec3 RankColor = vec3(0.5,0.5,0.5);
	switch (QueueID) {
		case 2:
		{
			DIVISION_COLOR.Get(tostring(position), RankColor);
			break;
		}
		case 3:
		{
			ROYAL_DIVISION_COLOR.Get(tostring(position), RankColor);
			break;
		}
	}
	return RankColor;
}

string GetDivisionText(uint position, uint QueueID) {
	string RankTitle = "[-]";
	vec3 RankColor = vec3(0.5,0.5,0.5);
	switch (QueueID) {
		case 2:
		{
			DIVISION_NAME.Get(tostring(position), RankTitle);
			DIVISION_COLOR.Get(tostring(position), RankColor);
			break;
		}
		case 3:
		{
			ROYAL_DIVISION_NAME.Get(tostring(position), RankTitle);
			ROYAL_DIVISION_COLOR.Get(tostring(position), RankColor);
			break;
		}
		case 4:
		{
			return "Super Royal";
		}
	}
	return CreateColorCode(RankColor) + RankTitle;
}

string GetDivisionTextShort(uint position, uint QueueID) {
	string RankTitle = "[-]";
	vec3 RankColor = vec3(0.5,0.5,0.5);
	switch (QueueID) {
		case 2:
		{
			DIVISION_SHORT.Get(tostring(position), RankTitle);
			DIVISION_COLOR.Get(tostring(position), RankColor);
			break;
		}
		case 3:
		{
			ROYAL_DIVISION_SHORT.Get(tostring(position), RankTitle);
			ROYAL_DIVISION_COLOR.Get(tostring(position), RankColor);
			break;
		}
		case 4:
		{
			return "";
		}
	}
	return CreateColorCode(RankColor) + RankTitle;
}

UI::Texture@ GetDivisionImage(uint position, uint QueueID, bool isPreview) {
	switch (QueueID) {
		case 2:
		{
			return G_UIState.MMTextures.Exists(tostring(position)) ? cast<UI::Texture>(G_UIState.MMTextures[tostring(position)]) : null;
		}
		case 3:
		{
			if (isPreview) {
				return G_UIState.RoyalPreviews.Exists(tostring(position)) ? cast<UI::Texture>(G_UIState.RoyalPreviews[tostring(position)]) : null;
			} else {
				return G_UIState.RoyalTextures.Exists(tostring(position)) ? cast<UI::Texture>(G_UIState.RoyalTextures[tostring(position)]) : null;
			}
		}
		case 4:
		{
			if (isPreview) {
				return G_UIState.RoyalTextures.Exists(tostring(0)) ? cast<UI::Texture>(G_UIState.RoyalTextures[tostring(0)]) : null;
			} else {
				return G_UIState.SuperRoyalTexture;
			}
		}
	}
	return null;
}

void RefreshRankColor() {
	vec3 rankColor = GetDivisionColor(G_QueueState.CurrentQueue.Stats.CurrentDivision.Position, G_QueueState.CurrentQueue.QueueID);
	G_UIState.RankColor.SetColor(rankColor);
	G_UIState.TabBarBottom = vec4(rankColor.x - 0.1,rankColor.y - 0.1, rankColor.z - 0.1, 0.7);
}

void RequestNextLeaderboardPage() {
	while (G_QueueState.CurrentQueue.Stats.Leaderboard.LastRequestStamp + 2000 > Time::Now) yield();
	G_QueueState.CurrentQueue.Stats.Leaderboard.RequestNextPage();
}

CTrackManiaPlayerInfo@ GetMainUserInfo() {
	auto app = cast<CTrackMania>(GetApp());
	return cast<CTrackManiaPlayerInfo>(cast<CTrackManiaNetwork>(app.Network).PlayerInfo);
}

void GenerateNewPartyCode() {
	string PartyCodeResponse = SendNadeoRequest(Net::HttpMethod::Post, "https://meet.trackmania.nadeo.club/api/code");
	Json::Value PartyCodeJSON = ResponseToJSON(PartyCodeResponse, Json::Type::Object);

	if (PartyCodeJSON.GetType() == Json::Type::Object) {
		try {
			G_DataState.PartyCode = PartyCodeJSON["code"];
		} catch {
			Error(ErrorType::Error, "GenerateNewPartyCode", "Couldn't generate new party code!", getExceptionInfo() + "\n" + PartyCodeResponse);
		}
	}
}

QueueRanking@ RefreshQueueRanking(uint QueueID) {
	QueueRanking@ Ranking = QueueRanking();

	string LeaderboardStatusResponse = SendNadeoRequest(Net::HttpMethod::Get, "https://meet.trackmania.nadeo.club/api/matchmaking/" + QueueID + "/leaderboard/players?players[]=" + GetMainUserInfo().WebServicesUserId);
	Json::Value LeaderboardStatus = ResponseToJSON(LeaderboardStatusResponse, Json::Type::Object);

	if (LeaderboardStatus.GetType() != Json::Type::Null) {
		try {
			if (uint(LeaderboardStatus["matchmakingId"]) == QueueID) {
				if (LeaderboardStatus["results"].Length == 1) {
					Ranking.Rank = LeaderboardStatus["results"][0]["rank"];
					Ranking.Points = LeaderboardStatus["results"][0]["score"];
				} else if (LeaderboardStatus["results"].Length == 0) {
					const array<string> QueueIDsForError = {"Teams 3v3", "Royal", "Super Royal"};
					Error(ErrorType::Log, "RefreshQueueRanking", "Couldn't get your Leaderboard status for Queue " + QueueID + (QueueID > 1 && QueueID < 5 ? (" (" + QueueIDsForError[QueueID - 2] + ")") : "") + ". If you have never queued for it, that's most likely the reason why this happened.", LeaderboardStatusResponse);
				} else {
					Error(ErrorType::Error, "RefreshQueueRanking", "The nadeo API responded with multiple results for your leaderboard status!", LeaderboardStatusResponse);
				}
			} else {
				Error(ErrorType::Warn, "RefreshQueueRanking", "LeaderboardStatus returned a different matchmaking ID than requested!", LeaderboardStatusResponse);
			}
			Ranking.Cardinal = LeaderboardStatus["cardinal"];
		} catch {
			Error(ErrorType::Warn, "RefreshQueueRanking", "Something went wrong while processing the Leaderboard Status!", getExceptionInfo() + "\n" + "Queue ID: " + QueueID + "\n" + LeaderboardStatusResponse);
		}
	} else {
		Error(ErrorType::Warn, "RefreshQueueRanking", "Leaderboard Status returned a bad response!", LeaderboardStatusResponse);
	}

	string ProgressionStatusResponse = SendNadeoRequest(Net::HttpMethod::Get, "https://meet.trackmania.nadeo.club/api/matchmaking/" + QueueID + "/progression/players?players[]=" + GetMainUserInfo().WebServicesUserId);
	Json::Value ProgressionStatus = ResponseToJSON(ProgressionStatusResponse, Json::Type::Object);

	if (ProgressionStatus.GetType() != Json::Type::Null) {
		try {
			if (ProgressionStatus.HasKey("progressions") && ProgressionStatus["progressions"].GetType() == Json::Type::Array) {
				if (ProgressionStatus["progressions"].Length == 1) {
					Ranking.Wins = ProgressionStatus["progressions"][0]["progression"];
				} else {
					Error(ErrorType::Log, "RefreshQueueRanking", "ProgressionStatus returned a different amount of players than requested!\n" + (ProgressionStatus["results"].Length != 0 ? "(Got rankings for " + ProgressionStatus["progressions"].Length + " players)!" : "This can happen if you have never queued up for a mode."), ProgressionStatusResponse);
				}
			} else {
				Error(ErrorType::Warn, "RefreshQueueRanking", "\"progressions\" not found in JSON", ProgressionStatusResponse);
			}
		} catch {
			Error(ErrorType::Warn, "RefreshQueueRanking", "Something went wrong while processing the Progression Status. This can happen if you have never queued up for a mode.", getExceptionInfo() + "\n" + "Queue ID: " + QueueID + "\n" + ProgressionStatusResponse);
		}
	} else {
		Error(ErrorType::Warn, "RefreshQueueRanking", "Progression Status returned a bad response! This can happen if you have never queued up for a mode.", LeaderboardStatusResponse);
	}

	return Ranking;
}

// --- Control Flow

void Error(ErrorType errType, const string &in context, const string &in message, const string &in log) {
	if (errType == ErrorType::Warn) {
		warn("[!] WARN [!]");
		warn("CONTEXT: " + context);
		warn("MESSAGE: " + message);
		warn("LOG: " + log);
		if (!Setting_HideWarn) UI::ShowNotification("\\$000" + Icons::ExclamationTriangle + " Warning in " + context, "\\$444" + message, vec4(1.0, 1.0, 0.0, 0.27), 5000);
	} else if (errType == ErrorType::Error) {
		error("[!] ERROR [!]");
		error("CONTEXT: " + context);
		error("MESSAGE: " + message);
		error("LOG: " + log);
		UI::ShowNotification(Icons::Times + " Error in " + context, message, vec4(1.0, 0.0, 0.0, 0.27), 5000);
	} else if (errType == ErrorType::UserWarning) {
		UI::ShowNotification(Icons::ExclamationTriangle + " " + message, log, vec4(0.6, 0.6, 1.0, 0.27), 5000);
	} else if (errType == ErrorType::Info) {
		UI::ShowNotification(Icons::Info + " " + message, log, vec4(0.6, 0.6, 1.0, 0.27), 5000);
	} else if (errType == ErrorType::Log) {
		trace(context + ": " + message + " (" + log + ")");
	}
}

void PlayWarnSound() {
	PlaySound(Sounds::Warn);
}

void PlayErrorSound() {
	PlaySound(Sounds::Error);
}

void PlayInfoSound() {
	PlaySound(Sounds::Info);
}

void PlayInviteSound() {
	PlaySound(Sounds::InvitationReceived);
}

void PlayMatchFoundSound() {
	PlaySound(Sounds::MatchFound);
}

void PlayMatchFoundInputSound() {
	PlaySound(Sounds::MatchFoundAwaitInput);
}

void PlayMatchFoundInputReceivedSound() {
	PlaySound(Sounds::MatchFoundInputReceived);
}

void PlayRankUpSound() {
	PlaySound(Sounds::RankUp);
}

void PlayRankDownSound() {
	PlaySound(Sounds::RankDown);
}

// --- Input plugin functions

void OnKeyPress(bool down, VirtualKey key) {
	if (G_DataState is null) return;
	G_DataState.LastActivityTick = Time::Now;
	G_DataState.LastActivityType = 0;
}

void OnMouseButton(bool down, int button, int x, int y) {
	if (G_DataState is null) return;
	G_DataState.LastActivityTick = Time::Now;
	G_DataState.LastActivityType = 1;
}

void OnMouseMove(int x, int y) {
	if (G_DataState is null) return;
	G_DataState.LastActivityTick = Time::Now;
	G_DataState.LastActivityType = 1;
	return;
}

// --- Misc

string GetPlacementFromRank(int rank) {
	if (rank < 1) return "" + rank;
	switch (rank % 10) {
		case 1 :
			if (rank % 100 == 11) return "" + rank + "th";
			return "" + rank + "st";
		case 2 :
			if (rank % 100 == 12) return "" + rank + "th";
			return "" + rank + "nd";
		case 3 :
			if (rank % 100 == 13) return "" + rank + "th";
			return "" + rank + "rd";
	}
	return "" + rank + "th";
}

bool CompareUplayFriends(const UplayFriend@ const &in a, const UplayFriend@ const &in b) {
	if (a.Presence != b.Presence) return (a.Presence == "Online" || b.Presence == "Offline");
	else return a.UplayPlayer.DisplayName.ToLower() < b.UplayPlayer.DisplayName.ToLower();
}

bool CompareRecentMembers(const RecentTeammate@ const &in a, const RecentTeammate@ const &in b) {
	return a.LastQueueStamp < b.LastQueueStamp;
}

namespace Math {
	vec3 Fract(const vec3 &in c) {
		return vec3(Math::Fract(c.x),Math::Fract(c.y),Math::Fract(c.z));
	}

	vec3 Clamp(const vec3 &in c, const float min, const float max) {
		return vec3(Math::Clamp(c.x,min,max),Math::Clamp(c.y,min,max),Math::Clamp(c.z,min,max));
	}

	vec3 Abs(const vec3 &in c) {
		return vec3(Math::Abs(c.x),Math::Abs(c.y),Math::Abs(c.z));
	}

	float Fract(const float v) {
		return v - Math::Floor(v);
	}
}